#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_relative_pose_evaluation
# just install "pip install cnspy_relative_pose_evaluation"
########################################################################################################################

import rosbag
import time
import os
import argparse
import yaml
import csv
from tqdm import tqdm
import numpy as np
from numpy import linalg as LA
from spatialmath import UnitQuaternion, SO3, SE3, Quaternion, base

from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from std_msgs.msg import Header, Time
from geometry_msgs.msg import PoseStamped, PoseWithCovarianceStamped, TransformStamped

from cnspy_trajectory.BsplineSE3 import TrajectoryInterpolationType
from cnspy_relative_pose_evaluation.ROSBag_Poses2RelPoses import *
from cnspy_relative_pose_evaluation.ROSBag_TrueRelPoses import *
from cnspy_relative_pose_evaluation.RelPose_ROSBag2CSV import *
from cnspy_relative_pose_evaluation.AssociateRelPoses import AssociateRelPoses, AssociateRelPoseCfg
from cnspy_relative_pose_evaluation.RelPoseMeasEvaluation import RelPoseMeasEvaluation


class RelPoseMeasEvaluationTool:
    @staticmethod
    def evaluate(bagfile_in:str,
                 cfg_fn:str,
                 result_dir=None,
                 verbose=True,
                 show_plot=True,
                 save_plot=True,
                 ID_arr=None,
                 filter_histogram=True,
                 extra_plots=True,
                 max_range=None,
                 max_angle=None,
                 remove_outliers=True,
                 interp_type: TrajectoryInterpolationType = TrajectoryInterpolationType.linear,
                 min_dt=0.05,
                 pose_error_type: EstimationErrorType = EstimationErrorType.type5):
        if not os.path.isfile(bagfile_in):
            print("RangeEvaluationTool: could not find file: %s" % bagfile_in)
            return False

        ## create result dir:
        if result_dir == "" or result_dir is None:
            folder = str(bagfile_in).replace(".bag", "")
        else:
            folder = result_dir

        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass
        result_dir = folder
        if verbose:
            print("* result_dir: \t " + str(folder))

        dict_cfg = None
        with open(cfg_fn, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "true_pose_topics" not in dict_cfg:
                print("[true_pose_topics] does not exist in fn=" + cfg_fn)
                return False
            if "relpose_topics" not in dict_cfg:
                print("[relpose_topics] does not exist in fn=" + cfg_fn)
                return False
            # if "new_relpose_topics" not in dict_cfg:
            #    print("[new_relpose_topics] was specified and is currently NOT SUPPORTED (please remove or comment it out) in fn=" + cfg_fn)
            #    return False
            print("Successfully read YAML config file.")

        Sensor_ID_arr = []
        Object_ID_arr = []
        topic_list = []
        for key, val in dict_cfg["true_pose_topics"].items():
            topic_list.append(val)
        for key, val in dict_cfg["relpose_topics"].items():
            topic_list.append(val)
            Sensor_ID_arr.append(int(key))
        if "object_positions" in dict_cfg:
            for key, val in dict_cfg["object_positions"].items():
                Object_ID_arr.append(int(key))

        if ID_arr:
            Sensor_ID_arr = ID_arr

        # unique IDs
        Sensor_ID_arr = list(set(Sensor_ID_arr))
        Object_ID_arr = list(set(Object_ID_arr))
        if verbose:
            print("* topic_list= " + str(topic_list))
            print("* Sensor_ID_arr= " + str(Sensor_ID_arr))
            print("* Object_ID_arr= " + str(Sensor_ID_arr))

        # topic_list=['/d01/ranging', '/a01/ranging', '/a02/ranging', '/a03/ranging']
        fn_meas_ranges = str(result_dir + '/all-meas-ranges.csv')
        fn_gt_ranges = str(result_dir + '/all-true-ranges.csv')

        IDs_found = None
        # 1) extract all measurements to CSV
        if not os.path.isfile(fn_meas_ranges):
            res, IDs_found = RelPose_ROSBag2CSV.extract_to_one(bagfile_name=bagfile_in,
                                                    cfg=cfg_fn,
                                                    fn=fn_meas_ranges,
                                                    result_dir=result_dir,
                                                    verbose=verbose,
                                                    with_cov=True)

        bagfile_out = str(result_dir + '/true-ranges.bag')

        # 2) create a clean bag file
        if not os.path.isfile(bagfile_out):
            res= ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                              bagfile_out_name=bagfile_out,
                                              cfg=cfg_fn,
                                              verbose=verbose,
                                              use_header_timestamp=False,
                                              ignore_new_topic_name=True,
                                              stddev_pos=0.0,
                                              interp_type=interp_type,
                                              min_dt=min_dt)

        if not os.path.isfile(fn_gt_ranges) or IDs_found is None:
            # 3) extract all measurements from the clean bagfile
            res, IDs_found = RelPose_ROSBag2CSV.extract_to_one(bagfile_name=bagfile_out,
                                                    cfg=cfg_fn,
                                                    fn=fn_gt_ranges,
                                                    result_dir=result_dir,
                                                    verbose=verbose)

        # 4) evaluate the ranges
        cfg = AssociateRelPoseCfg(ID1=None,
                                  ID2=None,
                                  relative_timestamps=False,
                                  max_difference=10 ** -4,
                                  subsample=0,
                                  verbose=verbose,
                                  remove_outliers=remove_outliers,
                                  pose_error_type=pose_error_type,
                                  range_error_val=0,
                                  label_timestamp='t',
                                  label_ID1='ID1',
                                  label_ID2='ID2',
                                  label_range='range')

        if max_range is not None:
            cfg.max_range_error = max_range
            cfg.max_range = max_range
        if max_angle is not None:
            cfg.max_angle = max_angle

        if IDs_found is not None:
            # remove IDs from Sensor_ID_arr, if they are not part of the bagfile!
            IDs = list()
            for ID in Sensor_ID_arr:
                if ID in IDs_found:
                    IDs.append(ID)
                elif verbose:
                    print("* Sensor ID= " + str(ID) + "was not found in the rosbag!" )
            Sensor_ID_arr = IDs

        eval = RelPoseMeasEvaluation(fn_gt=fn_gt_ranges,
                                     fn_est=fn_meas_ranges,
                                     ID1_arr=Sensor_ID_arr,
                                     ID2_arr=list(set(list(Sensor_ID_arr + Object_ID_arr))),
                                     cfg=cfg,
                                     result_dir=str(result_dir + '/eval/'),
                                     prefix='',
                                     save_plot=save_plot,
                                     show_plot=show_plot,
                                     save_statistics=True,
                                     plot_timestamps=extra_plots,
                                     plot_ranges=extra_plots,
                                     plot_angles=extra_plots,
                                     plot_ranges_sorted=extra_plots,
                                     plot_range_error=extra_plots,
                                     plot_angle_error=extra_plots,
                                     plot_range_histogram=True,
                                     plot_angle_histogram=True,
                                     filter_histogram=filter_histogram,
                                     plot_position_err=True,
                                     plot_pose_err=True,
                                     plot_pose=True,
                                     verbose=verbose
                                     )
        pass  # DONE


def main():
    # RelPoseMeasEvaluationTool.py --bagfile  ./test/sample_data/T1_A3_loiter_2m_2023-08-31-20-58-20.bag --cfg ./test/sample_data/config.yaml
    parser = argparse.ArgumentParser(
        description='RelPoseMeasEvaluationTool: evaluation the measured relative poses')
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--bagfile', help='input bag file', default="not specified")
    parser.add_argument('--cfg',
                        help='YAML configuration file describing the setup: ' +
                             '{sensor_positions:{<id>:[x,y,z], ...}, sensor_orientations:{<id>:[w,x,y,z], ...}, ' +
                             'relpose_topics:{<id>:<topic_name>, ...}, true_pose_topics:{<id>:<topic_name>, ...}',
                        default="config.yaml", required=True)
    parser.add_argument('--save_plot', help='saves all plots in the result_dir', action='store_true', default=False)
    parser.add_argument('--show_plot',
                        help='blocks the evaluation, for inspecting individual plots, continuous after closing',
                        action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--extra_plots', help='plots: timestamps, ranges + angles (sorted + unsorted + error),  ',
                        action='store_true', default=False)
    parser.add_argument('--keep_outliers', help='do not apply the max. thresholds on the error', action='store_true',
                        default=False)
    parser.add_argument('--filter_histogram',
                        help='filters the error histogram, such that the fitted normal distribution is computed on the best bins only',
                        action='store_true', default=False)
    parser.add_argument('--max_range', help='max. range that classifies them as outlier (0 disables feature). ',
                        default='0')
    parser.add_argument('--max_angle', help='max. range that classifies them as outlier (0 disables feature)',
                        default='0')
    parser.add_argument('--interpolation_type', help='Trajectory interpolation type',
                        choices=TrajectoryInterpolationType.list(),
                        default=str(TrajectoryInterpolationType.linear))
    parser.add_argument('--min_dt',
                        help='temporal displacement of cubic spline control points',
                        default=0.05)
    parser.add_argument('--pose_error_type', help='Covariance perturbation type (space) of relative pose measurements',
                        choices=EstimationErrorType.list(),
                        default=str(EstimationErrorType.type5))

    tp_start = time.time()
    args = parser.parse_args()

    RelPoseMeasEvaluationTool.evaluate(bagfile_in=args.bagfile,
                                       cfg_fn=args.cfg,
                                       result_dir=args.result_dir,
                                       save_plot=args.save_plot,
                                       show_plot=args.show_plot,
                                       verbose=args.verbose,
                                       max_range=float(args.max_range),
                                       max_angle=float(args.max_angle),
                                       extra_plots=args.extra_plots,
                                       filter_histogram=args.filter_histogram,
                                       remove_outliers=not args.keep_outliers,
                                       interp_type=TrajectoryInterpolationType(args.interpolation_type),
                                       min_dt=float(args.min_dt),
                                       pose_error_type=EstimationErrorType(args.pose_error_type)
                                       )
    pass
    print(" ")
    print("finished after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Requirements:
# os
########################################################################################################################
import os
from sys import version_info
from datetime import datetime
import argparse
import time
import math
import yaml
import pandas as pandas
import numpy as np

import cnspy_numpy_utils.numpy_statistics
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_timestamp_association.TimestampAssociation import TimestampAssociation
from cnspy_relative_pose_evaluation.AssociateRelPoses import AssociateRelPoses, AssociateRelPoseCfg
from matplotlib import pyplot as plt

from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig


class RelPoseMeasEvaluation:
    report = None

    def __init__(self,
                 fn_gt,
                 fn_est,
                 ID1_arr=[0], ID2_arr=[1],
                 result_dir=None,
                 prefix=None,
                 save_plot=True,
                 save_statistics=True,
                 show_plot=True,
                 cfg=AssociateRelPoseCfg(),
                 plot_timestamps=True,
                 plot_ranges=True,
                 plot_angles=True,
                 plot_ranges_sorted=True,
                 plot_range_error=True,
                 plot_angle_error=True,
                 plot_range_histogram=True,
                 plot_angle_histogram=True,
                 plot_pose_err=True,
                 plot_position_err=True,
                 plot_pose=True,
                 plot_NEES=True,
                 verbose=False,
                 filter_histogram=True,
                 ):
        if not result_dir:
            result_dir = '.'
        if not prefix:
            prefix = ''

        statistics_file = None
        if save_statistics:
            if not os.path.exists(result_dir):
                os.makedirs(result_dir)
            stat_fn = os.path.join(result_dir, 'statistics.yaml')
            statistics_file = open(stat_fn, 'w')
            if verbose:
                print("RelPoseMeasEvaluation: stat_fn=" + stat_fn)
            yaml.dump({'info': 'RelPoseMeasEvaluation Statistics',
                       'time': datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
                       'fn_gt': fn_gt,
                       'fn_est': fn_est,
                       'filter histogram': filter_histogram,
                       'remove outliers': cfg.remove_outliers,
                       'outliers max range': cfg.max_range,
                       'pose error type': str(cfg.pose_error_type),
                       'outliers min range': 0,
                       'outliers max angle': cfg.max_angle,
                       'result_dir': result_dir}, statistics_file, sort_keys=False, explicit_start=False,
                      default_flow_style=False)
            yaml.dump({'ID1s': ID1_arr,
                       'ID2s': ID2_arr}, statistics_file, sort_keys=False, explicit_start=False,
                      default_flow_style=True)
        fn_gt = os.path.abspath(fn_gt)
        fn_est = os.path.abspath(fn_est)

        plt.style.use('ggplot')

        SMALL_SIZE = 6
        MEDIUM_SIZE = 7
        BIGGER_SIZE = 8

        plt.rc('font', size=SMALL_SIZE)  # controls default text sizes
        plt.rc('axes', titlesize=SMALL_SIZE)  # fontsize of the axes title
        plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

        for ID1 in ID1_arr:
            if plot_timestamps:
                fig_t = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_t.suptitle('Timestamps of ID=' + str(ID1), fontsize=16)
            if plot_ranges:
                fig_r = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_r.suptitle('Ranges of ID=' + str(ID1), fontsize=16)
            if plot_angles:
                fig_a = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_a.suptitle('Angles of ID=' + str(ID1), fontsize=16)
            if plot_ranges_sorted:
                fig_rs = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_rs.suptitle('Range sorted of ID=' + str(ID1), fontsize=16)
            if plot_range_error:
                fig_re = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_re.suptitle('Range Error of ID=' + str(ID1), fontsize=16)
            if plot_angle_error:
                fig_ae = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_ae.suptitle('Angle Error of ID=' + str(ID1), fontsize=16)
            if plot_position_err:
                fig_pos_err_dict = dict()
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_pe = plt.figure(figsize=(20, 15), dpi=int(200))
                    fig_pe.suptitle('Position Error of ID=' + str(ID1) + " to " + str(ID2), fontsize=16)
                    fig_pos_err_dict[ID2] = fig_pe
            if plot_pose_err:
                fig_pe_dict = dict()
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_pe = plt.figure(figsize=(20, 15), dpi=int(200))
                    fig_pe.suptitle('Pose Error of ID=' + str(ID1) + " to " + str(ID2), fontsize=16)
                    fig_pe_dict[ID2] = fig_pe
            if plot_pose:
                fig_p_dict = dict()
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_p = plt.figure(figsize=(20, 15), dpi=int(200))
                    fig_p.suptitle('Pose of ID=' + str(ID1) + " to " + str(ID2), fontsize=16)
                    fig_p_dict[ID2] = fig_p
            if plot_NEES:
                fig_n = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_n.suptitle('NEES of ID=' + str(ID1), fontsize=16)
            if plot_range_histogram:
                fig_hr = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_hr.suptitle('Range Error Histograms of ID=' + str(ID1), fontsize=16)
            if plot_angle_histogram:
                fig_ha = plt.figure(figsize=(20, 15), dpi=int(200))
                fig_ha.suptitle('Angle Error Histograms of ID=' + str(ID1), fontsize=16)
            if save_statistics:
                dict_statistics_i = {'ID': ID1,
                                     'range_constant_bias_table': dict(),
                                     'range_noise_table': dict(),
                                     'angle_constant_bias_table': dict(),
                                     'angle_noise_table': dict(),
                                     'percent_outliers' : dict(),
                                     'ARMSE_p': dict(),
                                     'ARMSE_q_deg': dict(),
                                     'scale': dict(),
                                     'samples': dict(),
                                     }
                pass


            n = len(ID2_arr)
            if ID1 in ID2_arr:
                n = n - 1

            sqrt_n = max(1, math.ceil(math.sqrt(n)))
            n_rows = sqrt_n
            if sqrt_n * sqrt_n < n:
                n_cols = sqrt_n + 1
            else:
                n_cols = sqrt_n

            if n_rows * n_cols <= n:
                assert(False, "something went wrong!")

            idx = 1
            for ID2 in ID2_arr:
                if ID1 == ID2:
                    continue

                cfg.ID1 = int(ID1)
                cfg.ID2 = int(ID2)
                cfg_title = str("ID" + str(ID1) + " to ID" + str(ID2))
                assoc = AssociateRelPoses(fn_gt=fn_gt, fn_est=fn_est, cfg=cfg)
                if not assoc.data_loaded:
                    print("WARING: no measurement for " + cfg_title + "found!")
                    continue
                assoc.save(result_dir=result_dir, prefix=prefix + cfg_title)

                if save_statistics:
                    dict_statistics_i['percent_outliers'][ID2] = round(float(assoc.perc_outliers), 2)
                    dict_nees = assoc.avg_NEES()
                    for key, val in dict_nees.items():
                        if not (key in dict_statistics_i):
                            dict_statistics_i[key] = dict()
                        dict_statistics_i[key][ID2] = round(float(val),2)
                    ARMSE_p, ARMSE_q_deg = assoc.traj_err.get_ARMSE()
                    dict_statistics_i['ARMSE_p'][ID2] = round(float(ARMSE_p),2)
                    dict_statistics_i['ARMSE_q_deg'][ID2] = round(float(ARMSE_q_deg),2)
                    dict_statistics_i['scale'][ID2] = round(float(assoc.traj_err.scale),2)
                    dict_statistics_i['samples'][ID2] = assoc.num_samples()
                if plot_timestamps:
                    ax_t = fig_t.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_timestamps(fig=fig_t, ax=ax_t, calc_error=True, cfg_title=cfg_title)

                if plot_ranges:
                    ax_r = fig_r.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_ranges(fig=fig_r, ax=ax_r, cfg_title=cfg_title)

                if plot_angles:
                    ax_a = fig_a.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_angles(fig=fig_a, ax=ax_a, cfg_title=cfg_title)
                if plot_angle_error:
                    ax_ae = fig_ae.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_angle_error(fig=fig_ae, ax=ax_ae, cfg_title=cfg_title)

                if plot_ranges_sorted:
                    ax_rs = fig_rs.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_ranges(fig=fig_rs, ax=ax_rs, sorted=True, cfg_title=cfg_title)

                if plot_position_err:
                    fig_pe = fig_pos_err_dict[ID2]
                    cfg_plt = TrajectoryPlotConfig()
                    cfg_plt.title = cfg_title
                    cfg_plt.show = False
                    cfg_plt.close_figure = False
                    cfg_plt.unwrap = False
                    assoc.plot_position_err(fig=fig_pe, cfg=cfg_plt)

                if plot_pose_err:
                    fig_pe = fig_pe_dict[ID2]
                    cfg_plt = TrajectoryPlotConfig()
                    cfg_plt.title = cfg_title
                    cfg_plt.show = False
                    cfg_plt.close_figure = False
                    cfg_plt.unwrap = False
                    assoc.plot_traj_err(fig=fig_pe, cfg=cfg_plt)

                if plot_pose:
                    fig_pe = fig_p_dict[ID2]
                    cfg_plt = TrajectoryPlotConfig()
                    cfg_plt.title = cfg_title
                    cfg_plt.show = False
                    cfg_plt.close_figure = False
                    cfg_plt.unwrap = True
                    assoc.plot_traj(fig=fig_pe, cfg=cfg_plt)

                if plot_NEES:
                    ax_n = fig_n.add_subplot(n_rows, n_cols, idx)
                    assoc.plot_NEES(fig=fig_n, ax=ax_n, relative_time=True, cfg_title=cfg_title)


                if plot_range_error:
                    ax_e = fig_re.add_subplot(n_rows, n_cols, idx)
                    [fig_, ax_, stat, r_vec_err_] = assoc.plot_range_error(fig=fig_re, ax=ax_e,
                                                                           sorted=False,
                                                                           remove_outlier=True,
                                                                           cfg_title=cfg_title)
                    cnspy_numpy_utils.numpy_statistics.print_statistics(stat, desc=cfg_title + " error")
                if plot_range_histogram:
                    ax_hr = fig_hr.add_subplot(n_rows, n_cols, idx)
                    [fig_, ax_, stat, r_vec_err_] = assoc.plot_range_error_histogram(fig=fig_hr,
                                                                                     ax=ax_hr,
                                                                                     max_error=cfg.max_range_error,
                                                                                     filter_histogramm=filter_histogram,
                                                                                     ID1=ID1, ID2=ID2)
                    if stat is not None and save_statistics:
                        dict_statistics_i['range_constant_bias_table'][ID2] = round(float(stat['mean']), 2)
                        dict_statistics_i['range_noise_table'][ID2] = round(float(stat['std']), 2)
                if plot_angle_histogram:
                    ax_ha = fig_ha.add_subplot(n_rows, n_cols, idx)
                    [fig_, ax_, stat, r_vec_err_] = assoc.plot_angle_error_histogram(fig=fig_ha,
                                                                                     ax=ax_ha,
                                                                                     max_error=10,
                                                                                     filter_histogramm=filter_histogram,
                                                                                     ID1=ID1, ID2=ID2)
                    if stat is not None  and save_statistics:
                        dict_statistics_i['angle_constant_bias_table'][ID2] = round(float(stat['mean']), 2)
                        dict_statistics_i['angle_noise_table'][ID2] = round(float(stat['std']), 2)
                # the histogram of the date
                idx += 1

            # for ID2
            if verbose:
                print("* RangeEvaluation(): ranges associated!")

            # Tweak spacing to prevent clipping of ylabel
            if plot_timestamps:
                fig_t.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_t, result_dir=result_dir,
                                                       save_fn=str("Timestamps" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)

            if plot_ranges:
                fig_r.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_r, result_dir=result_dir,
                                                       save_fn=str("Ranges_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_angles:
                fig_a.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_a, result_dir=result_dir,
                                                       save_fn=str("Angle_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_ranges_sorted:
                fig_rs.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_rs, result_dir=result_dir,
                                                       save_fn=str("Range_Sorted_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)

            if plot_range_error:
                fig_re.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_re, result_dir=result_dir,
                                                       save_fn=str("Range_Errors_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_position_err:
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_pe = fig_pos_err_dict[ID2]
                    fig_pe.tight_layout()
                    if save_plot:
                        AssociateRelPoses.show_save_figure(fig=fig_pe, result_dir=result_dir,
                                                           save_fn=str("Position_Errors_ID" + str(ID1) + "_to_" + str(ID2)),
                                                           show=show_plot, close_figure=not show_plot)
                    pass
            if plot_pose_err:
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_pe = fig_pe_dict[ID2]
                    fig_pe.tight_layout()
                    if save_plot:
                        AssociateRelPoses.show_save_figure(fig=fig_pe, result_dir=result_dir,
                                                           save_fn=str("Pose_Errors_ID" + str(ID1) + "_to_" + str(ID2)),
                                                           show=show_plot, close_figure=not show_plot)
                    pass

            if plot_pose:
                for ID2 in ID2_arr:
                    if ID1 == ID2:
                        continue
                    fig_pe = fig_p_dict[ID2]
                    fig_pe.tight_layout()
                    if save_plot:
                        AssociateRelPoses.show_save_figure(fig=fig_pe, result_dir=result_dir,
                                                           save_fn=str("Pose_ID" + str(ID1) + "_to_" + str(ID2)),
                                                           show=show_plot, close_figure=not show_plot)
                    pass
            if plot_angle_error:
                fig_ae.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_ae, result_dir=result_dir,
                                                       save_fn=str("Angle_Errors_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_range_histogram:
                fig_hr.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_hr, result_dir=result_dir,
                                                       save_fn=str("Range_Error_Histograms_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_angle_histogram:
                fig_ha.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_ha, result_dir=result_dir,
                                                       save_fn=str("Angle_Error_Histograms_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if plot_NEES:
                fig_n.tight_layout()
                if save_plot:
                    AssociateRelPoses.show_save_figure(fig=fig_n, result_dir=result_dir,
                                                       save_fn=str("NEES_ID" + str(ID1)),
                                                       show=show_plot, close_figure=not show_plot)
            if save_statistics:
                yaml.dump(dict_statistics_i, statistics_file, explicit_start=True, default_flow_style=True)
                if verbose:
                    print(yaml.dump(dict_statistics_i, explicit_start=True, default_flow_style=True))
            pass  # UWB_ID2
        pass  # UWB_ID1

        if save_statistics:
            statistics_file.close()

        pass  # DONE


def main():
    parser = argparse.ArgumentParser(
        description='RelPoseMeasEvaluation: evaluate and estimated and true pairwise relative pose measurements')
    parser.add_argument('--fn_gt', help='input ground-truth trajectory CSV file', default="not specified")
    parser.add_argument('--fn_est', help='input estimated  trajectory CSV file', default="not specified")
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--ID1s', help='ID of TX', nargs='+', default=[0])
    parser.add_argument('--ID2s', help='ID of RX', nargs='+', default=[1])
    parser.add_argument('--prefix', help='prefix in results', default='')
    parser.add_argument('--max_timestamp_difference',
                        help='Max difference between associated timestampes (t_gt - t_est)', default=0.03)
    parser.add_argument('--subsample', help='subsampling factor for input data (CSV)', default=0)
    parser.add_argument('--plot', action='store_true', default=True)
    parser.add_argument('--save_plot', action='store_true', default=True)
    parser.add_argument('--save_statistics', action='store_true', default=True)
    parser.add_argument('--show_plot', action='store_true', default=True)
    parser.add_argument('--relative_timestamps', action='store_true', default=False)
    parser.add_argument('--remove_outliers', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--max_range', help='range that classifies as outlier', default='30')
    parser.add_argument('--max_angle', help='angle that classifies as outlier', default='6.4')
    parser.add_argument('--range_error_val', help='value assigned to outlier', default='0')
    parser.add_argument('--angle_error_val', help='value assigned to outlier', default='0')
    parser.add_argument('--label_timestamp', help='timestamp label in CSV', default='t')
    parser.add_argument('--label_range', help='range label in CSV', default='range')
    parser.add_argument('--label_angle', help='range label in CSV', default='angle')
    parser.add_argument('--label_ID1', help='ID1 label in CSV', default='ID1')
    parser.add_argument('--label_ID2', help='ID2 label in CSV', default='ID2')
    parser.add_argument('--plot_timestamps', action='store_true', default=False)
    parser.add_argument('--plot_ranges', action='store_true', default=False)
    parser.add_argument('--plot_angles', action='store_true', default=False)
    parser.add_argument('--plot_err_poses', action='store_true', default=False)
    parser.add_argument('--plot_ranges_sorted', action='store_true', default=False)
    parser.add_argument('--plot_errors', action='store_true', default=False)
    parser.add_argument('--plot_histograms', action='store_true', default=False)
    parser.add_argument('--filter_histogram', action='store_true', default=False)
    tp_start = time.time()
    args = parser.parse_args()
    cfg = AssociateRelPoseCfg(ID1=None,
                              ID2=None,
                              relative_timestamps=args.relative_timestamps,
                              max_difference=float(args.max_timestamp_difference),
                              subsample=int(args.subsample),
                              verbose=args.verbose,
                              remove_outliers=args.remove_outliers,
                              max_range=float(args.max_range),
                              range_error_val=float(args.range_error_val),
                              label_timestamp=args.label_timestamp,
                              label_ID1=args.label_ID1,
                              label_ID2=args.label_ID2,
                              label_range=args.label_range)

    eval = RelPoseMeasEvaluation(fn_gt=args.fn_gt,
                                 fn_est=args.fn_est,
                                 ID1_arr=args.ID1s,
                                 ID2_arr=args.ID2s,
                                 cfg=cfg,
                                 result_dir=args.result_dir,
                                 prefix=args.prefix,
                                 save_plot=args.save_plot,
                                 show_plot=args.show_plot,
                                 save_statistics=args.save_statistics,
                                 plot_timestamps=args.plot_timestamps,
                                 plot_ranges=args.plot_ranges,
                                 plot_angles=args.plot_angles,
                                 plot_pose_err=args.plot_err_poses,
                                 plot_ranges_sorted=args.plot_ranges_sorted,
                                 plot_range_error=args.plot_errors,
                                 plot_angle_error=args.plot_errors,
                                 plot_range_histogram=args.plot_histograms,
                                 plot_angle_histogram=args.plot_histograms,
                                 filter_histogram=args.filter_histogram
                                 )

    print(" ")
    print("finished after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import unittest
import os

from cnspy_relative_pose_evaluation.RelPoseMeasEvaluation import *
from cnspy_relative_pose_evaluation.RelPoseMeasEvaluationTool import RelPoseMeasEvaluationTool
from cnspy_trajectory.BsplineSE3 import TrajectoryInterpolationType

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')
RES_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_results')


class RelPoseMeasEvaluation_Test(unittest.TestCase):
    def test_dataset_eval(self):
        DATASET_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/eval/"
        RES_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/eval/"

        RelPoseMeasEvaluationTool.evaluate(bagfile_in=SAMPLE_DATA_DIR + '/mukisano_6/output.bag',
                                           cfg_fn=SAMPLE_DATA_DIR + '/mukisano_6/config.yaml',
                                           result_dir=RES_DATA_DIR + '/mukisano_6',
                                           save_plot=True,
                                           show_plot=True,
                                           verbose=True,
                                           max_range=float(50),
                                           max_angle=None,
                                           extra_plots=True,
                                           filter_histogram=False,
                                           remove_outliers=False,
                                           ID_arr=[0],
                                           interp_type=TrajectoryInterpolationType.linear,
                                           min_dt=float(0.01),
                                           pose_error_type=EstimationErrorType.type1
                                           )



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import sys
import rosbag
import time
import os
import argparse
import yaml
from tqdm import tqdm
import numpy as np
from enum import Enum

from spatialmath import UnitQuaternion, SE3, base

from cnspy_trajectory.ROSBag_Pose import ROSBag_Pose
from cnspy_trajectory.HistoryBuffer import get_key_from_value
from cnspy_relative_pose_evaluation.ROSBag_TrueRelPoses import interpolate_pose, random_orientation
from geometry_msgs.msg import PoseWithCovarianceStamped


class PoseTypes(Enum):
    SE2 = 'SE2'
    SE3 = 'SE3'
    Pos3DYaw = 'Pos3DYaw'

    def __str__(self):
        return self.value

    @staticmethod
    def list():
        return list([str(PoseTypes.SE2),
                     str(PoseTypes.SE3),
                     str(PoseTypes.Pos3DYaw)])


class ROSBag_Pose2AbsPoses:
    def __init__(self):
        pass

    @staticmethod
    def extract(bagfile_in,
                bagfile_out,
                cfg,
                stddev_pos=0.0,
                stddev_or=0.0,
                use_header_timestamp=False,
                verbose=False,
                replace_with_new_topic=False
                ):

        if not os.path.isfile(bagfile_in):
            print("ROSBag_Pose2AbsPoses: could not find file: %s" % bagfile_in)
            return False
        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("ROSBag_Pose2AbsPoses: could not find file: %s" % cfg)
            return False

        if verbose:
            print("ROSBag_Pose2AbsPoses:")
            print("* bagfile in name: " + str(bagfile_in))
            print("* bagfile out name: " + str(bagfile_out))
            print("* cfg YAML file: \t " + str(cfg))
            print("* stddev_pos: " + str(stddev_pos))
            print("* stddev_or: " + str(stddev_or))
            print("* use_header_timestamp: " + str(use_header_timestamp))
            print("* replace_with_new_topic: " + str(replace_with_new_topic))

        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_in)
        except:
            if verbose:
                print("ROSBag_Pose2AbsPoses: Unexpected error!")

            return False

        ## create result dir:
        [root, ext] = os.path.splitext(bagfile_out)
        [head, tail] = os.path.split(root)
        try:  # else already exists
            os.makedirs(head)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(head))

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "sensor_positions" not in dict_cfg:
                print("[sensor_positions] does not exist in fn=" + cfg)
            if "sensor_orientations" not in dict_cfg:
                print("[sensor_orientations] does not exist in fn=" + cfg)
            if "sensor_topics" not in dict_cfg:
                print("[sensor_topics] does not exist in fn=" + cfg)
                return False
            if "true_pose_topics" not in dict_cfg:
                print("[true_pose_topics] does not exist in fn=" + cfg)
                return False
            if "new_sensor_topics" not in dict_cfg:
                print("[new_sensor_topics] does not exist in fn=" + cfg)
            if "pose_types" not in dict_cfg:
                print("[pose_types] does not exist in fn=" + cfg + " with types: " + str(PoseTypes.list()))
                return False
            print("Read successful")
        if verbose:
            print("configuration contains:")
            print("sensor_topics:" + str(dict_cfg["sensor_topics"]))
            print("true_pose_topics:" + str(dict_cfg["true_pose_topics"]))
            print("pose_types:" + str(dict_cfg["pose_types"]))
        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("ROSBag_Pose2AbsPoses: Unexpected error, bag file might be empty!")
            bag.close()
            return False

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']

        # check if desired topics are in
        all_found = True
        for id, topicName in dict_cfg["sensor_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
            if not found_:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")
                all_found = False
        for id, topicName in dict_cfg["true_pose_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
            if not found_:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")
                all_found = False

        if not all_found:
            print("Topics contained:")
            for t_ in bag_topics:
                print("* " + str(t_))
        if verbose:
            print("\nROSBag_Pose2AbsPoses: num messages " + str(num_messages))

        ## store the BODY SENSOR pose in a dictionary
        dict_T_BODY_SENSOR = dict() # map<topic, SE3>+
        for ID, sensor_topic in dict_cfg["sensor_topics"].items():
            t_BT = np.zeros((3, 1))
            if "sensor_positions" in dict_cfg and ID in dict_cfg["sensor_positions"].keys():
                t_BT = np.array(dict_cfg["sensor_positions"][ID])

            q_BT = UnitQuaternion()
            if "sensor_orientations" in dict_cfg and ID in dict_cfg["sensor_orientations"].keys():
                # expecting: w,x,y,z
                q_vec = np.array(dict_cfg["sensor_orientations"][ID])
                q_BT = UnitQuaternion(q_vec).unit()
                pass
            dict_T_BODY_SENSOR[dict_cfg["sensor_topics"][ID]] = SE3.Rt(q_BT.R, t_BT, check=True)

        ## extract the poses of the desired topics from the BAG file
        round_decimals = 6
        dict_history = ROSBag_Pose.extract_poses(bag=bag, num_messages=num_messages,
                                                 dict_topic_pose_body=dict_cfg["true_pose_topics"],
                                                 dict_senor_topic_pose=dict_cfg["sensor_topics"],
                                                 round_decimals=round_decimals,
                                                 dict_T_BODY_SENSOR=dict_T_BODY_SENSOR)
        if len(dict_history) == 0:
            return False

        cnt = 0
        try:  # else already exists
            print("ROSBag_Pose2AbsPoses: computing new absolute pose measurements of a certain type at given senor_topics")
            with rosbag.Bag(bagfile_out, 'w') as outbag:
                for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
                    if topic in dict_cfg["sensor_topics"].values():
                        id_topic = get_key_from_value(dict_cfg["sensor_topics"], topic)
                        timestamp = msg.header.stamp.to_sec()
                        msg_new = PoseWithCovarianceStamped()
                        msg_new.header = msg.header

                        pose_type = PoseTypes(dict_cfg["pose_types"][id_topic])
                        # check if the topic should be renamed.
                        topic_out = topic
                        if "new_sensor_topics" in dict_cfg:
                            if id_topic in dict_cfg["new_sensor_topics"]:
                                topic_out = dict_cfg["new_sensor_topics"][id_topic]

                                if not replace_with_new_topic:
                                    # store original message, otherwise it gets lost:
                                    if use_header_timestamp and hasattr(msg, "header"):
                                        outbag.write(topic, msg, msg.header.stamp)
                                    else:
                                        outbag.write(topic, msg, t)
                            else:
                                print("No new_sensor_topic with ID=[" + str(id_topic) + "] found!")
                                return False
                        T_GLOBAL_SENSOR1 = interpolate_pose(dict_history[topic], timestamp, round_decimals)
                        if T_GLOBAL_SENSOR1 is not None:
                            p = T_GLOBAL_SENSOR1.t
                            q = UnitQuaternion(T_GLOBAL_SENSOR1.R, norm=True).unit()
                            if stddev_pos > 0:
                                p_noise = np.random.normal(0, stddev_pos, size=3)
                                p = p + p_noise.T
                            if stddev_or > 0:
                                q_noise = random_orientation(stddev_rad=stddev_or)
                                q = q*q_noise
                            else:
                                pass
                            var_pos = stddev_pos * stddev_pos
                            var_or = stddev_or * stddev_or
                            if pose_type is PoseTypes.SE2:
                                rpy = base.tr2rpy(T=q.R, unit='rad', order='zyx')
                                rpy[0] = 0
                                rpy[1] = 0
                                R_ = base.rpy2r(rpy, unit='rad', order='zyx')
                                q = UnitQuaternion(R_)
                                qv = q.vec

                                msg_new.pose.pose.position.x = p[0]
                                msg_new.pose.pose.position.y = p[1]
                                msg_new.pose.pose.position.z = 0
                                msg_new.pose.pose.orientation.w = qv[0]
                                msg_new.pose.pose.orientation.x = qv[1]
                                msg_new.pose.pose.orientation.y = qv[2]
                                msg_new.pose.pose.orientation.z = qv[3]

                                if stddev_pos > 0:
                                    # upper left block
                                    for idx in [0, 7]:
                                        msg_new.pose.covariance[idx] = var_pos
                                if stddev_or > 0:
                                    # lower right block

                                    for idx in [35]:
                                        msg_new.pose.covariance[idx] = var_or
                            elif pose_type is PoseTypes.SE3:
                                qv = q.vec

                                msg_new.pose.pose.position.x = p[0]
                                msg_new.pose.pose.position.y = p[1]
                                msg_new.pose.pose.position.z = p[2]
                                msg_new.pose.pose.orientation.w = qv[0]
                                msg_new.pose.pose.orientation.x = qv[1]
                                msg_new.pose.pose.orientation.y = qv[2]
                                msg_new.pose.pose.orientation.z = qv[3]
                                if stddev_pos > 0:
                                    # upper left block
                                    var_pos = stddev_pos*stddev_pos

                                    for idx in [0, 7, 14]:
                                        msg_new.pose.covariance[idx] = var_pos
                                if stddev_or > 0:
                                    # lower right block
                                    var_or = stddev_or * stddev_or
                                    for idx in [21, 28, 35]:
                                        msg_new.pose.covariance[idx] = var_or
                            elif pose_type is PoseTypes.Pos3DYaw:
                                rpy = base.tr2rpy(T=q.R, unit='rad', order='zyx')
                                rpy[0] = 0
                                rpy[1] = 0
                                R_ = base.rpy2r(rpy, unit='rad', order='zyx')
                                q = UnitQuaternion(R_)
                                qv = q.vec

                                msg_new.pose.pose.position.x = p[0]
                                msg_new.pose.pose.position.y = p[1]
                                msg_new.pose.pose.position.z = p[2]
                                msg_new.pose.pose.orientation.w = qv[0]
                                msg_new.pose.pose.orientation.x = qv[1]
                                msg_new.pose.pose.orientation.y = qv[2]
                                msg_new.pose.pose.orientation.z = qv[3]

                                if stddev_pos > 0:
                                    # upper left block
                                    for idx in [0, 7, 14]:
                                        msg_new.pose.covariance[idx] = var_pos
                                if stddev_or > 0:
                                    # lower right block

                                    for idx in [35]:
                                        msg_new.pose.covariance[idx] = var_or
                            else:
                                print("ROSBag_Pose2AbsPoses: unsupported PoseTypes!")
                                return False
                            # for id2
                            if use_header_timestamp and hasattr(msg, "header"):
                                outbag.write(topic_out, msg_new, msg_new.header.stamp)
                            else:
                                outbag.write(topic_out, msg_new, t)
                            cnt += 1
                        pass
                    else:
                        # write other topic
                        if use_header_timestamp and hasattr(msg, "header"):
                            outbag.write(topic, msg, msg.header.stamp)
                        else:
                            outbag.write(topic, msg, t)
                        pass

        except  Exception as e:
            print("ROSBag_Pose2AbsPoses: Unexpected error while creating the bag file! msg=%s" % repr(e))
            print(str(sys.exc_info()))
            return False
        ## CLEANUP:
        except AssertionError as error:
            print(error)
            print("ROSBag_Pose2AbsPoses: Unexpected error while creating bag file")
            return False

        if verbose:
            print("\nROSBag_Pose2AbsPoses: " + str(cnt) + " measurements modified!")
        bag.close()
        return True

def main():
    # example: ROSBag_Pose2Ranges.py --bagfile ../test/sample_data//uwb_calib_a01_2023-08-31-21-05-46.bag --topic /d01/mavros/vision_pose/pose --cfg ../test/sample_data/config.yaml --verbose
    parser = argparse.ArgumentParser(
        description='ROSBag_Pose2AbsPoses: extract given pose topics and compute for each sensor topic poses a ' +
                    'absolute pose measurement (which can be perturbed) of a specific type which is added to a ' +
                    'new bag file')
    parser.add_argument('--bagfile_in', help='input bag file', required=True)
    parser.add_argument('--bagfile_out', help='output bag file', required=True)
    parser.add_argument('--cfg',
                        help='YAML configuration file describing the setup: ' +
                             '{sensor_positions:{<id>:[x,y,z], ...}, sensor_orientations:{<id>:[w,x,y,z], ...}, ' +
                             'sensor_topics:{<id>:<topic_name>, ...}, true_pose_topics:{<id>:<topic_name>, ...}, ' +
                             'new_sensor_topics:{<id>:<topic_name>, ...}, pose_types:{<id>:<SE2, SE3, or Pos3DYaw>, ...}',
                        default="config.yaml", required=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--std_pos',
                        help='standard deviation of generated measurements: z_p = p + white_noise(std_p)',
                        default=0.0)
    parser.add_argument('--std_or',
                        help='standard deviation of generated measurements: z_R = R *R(white_noise(std_or))',
                        default=0.0)
    parser.add_argument('--use_header_timestamp', action='store_true',
                        help='overwrites the bag time with the header time stamp', default=False)
    parser.add_argument('--replace_with_new_topic', action='store_true',
                        help='removes sensor_topic if a new_sensor_topic was specified', default=False)


    tp_start = time.time()
    args = parser.parse_args()

    if ROSBag_Pose2AbsPoses.extract(bagfile_in=args.bagfile_in,
                                    bagfile_out=args.bagfile_out,
                                    cfg=args.cfg,
                                    verbose=args.verbose,
                                    stddev_pos=float(args.std_pos),
                                    stddev_or=float(args.std_or),
                                    use_header_timestamp=args.use_header_timestamp,
                                    replace_with_new_topic=args.replace_with_new_topic):
        print(" ")
        print("finished after [%s sec]\n" % str(time.time() - tp_start))
    else:
        print("failed! after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass


#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import sys
import rosbag
import time
import os
import argparse
import yaml
import csv
from tqdm import tqdm
import numpy as np
from numpy import linalg as LA
from spatialmath import UnitQuaternion, SO3, SE3
from spatialmath.base.quaternions import qslerp
from cnspy_trajectory.HistoryBuffer import get_key_from_value
from cnspy_trajectory.ROSBag_Pose import ROSBag_Pose


def relpose_to_csv_line(timestamp, ID1, ID2, Pose, round_decimals=4):
    t = Pose.t
    range = LA.norm(t)
    q = UnitQuaternion(Pose.R)
    qv = q.vec
    [ang, vec] = q.angvec()
    content = ["%f" % round(timestamp, round_decimals), str(ID1), str(ID2),
               str(round(t[0], round_decimals)),
               str(round(t[1], round_decimals)),
               str(round(t[2], round_decimals)),
               str(round(qv[3], round_decimals)),
               str(round(qv[0], round_decimals)),
               str(round(qv[1], round_decimals)),
               str(round(qv[2], round_decimals)),
               str(round(range, round_decimals)),
               str(round(ang, round_decimals))]
    return content



class ROSBag_Poses2RelPoses:
    def __init__(self):
        pass

    @staticmethod
    def extract(bagfile_name, cfg, filename, result_dir="", verbose=False):

        if not os.path.isfile(bagfile_name):
            print("ROSBag_Poses2RelPoses: could not find file: %s" % bagfile_name)
            return False
        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("ROSBag_Poses2RelPoses: could not find file: %s" % cfg)
            return False

        if verbose:
            print("ROSBag_Poses2RelPoses:")
            print("* bagfile name: " + str(bagfile_name))
            print("* cfg: \t " + str(cfg))
            print("* filename: " + str(filename))

        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_name)
        except:
            if verbose:
                print("ROSBag_Poses2RelPoses: Unexpected error!")

            return False

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "sensor_positions" not in dict_cfg:
                print("[sensor_positions] does not exist in fn=" + cfg)
                return False
            if "sensor_orientations" not in dict_cfg:
                print("[sensor_orientations] does not exist in fn=" + cfg)
            if "true_pose_topics" not in dict_cfg:
                print("[true_pose_topics] does not exist in fn=" + cfg)
                return False
            print("Read successful")
        if verbose:
            print("configuration contains:")
            print("sensor_positions:" + str(dict_cfg["sensor_positions"]))
            print("sensor_orientations:" + str(dict_cfg["sensor_orientations"]))
            print("true_pose_topics:" + str(dict_cfg["true_pose_topics"]))

        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("ROSBag_Poses2RelPoses: Unexpected error, bag file might be empty!")
            bag.close()
            return False

        ## create result dir:
        if result_dir == "":
            folder = str(bagfile_name).replace(".bag", "")
        else:
            folder = result_dir

        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(folder))

        ## create csv file according to the topic names:
        dict_file_writers = dict()
        dict_header_written = dict()
        dict_csvfile_hdls = dict()
        idx = 0

        ## create a file writer for each sensor topic
        for  id, topicName in dict_cfg["true_pose_topics"].items():
            if filename is None:
                fn = str(folder + '/') + str.replace(topicName[1:], '/', '_') + '_sensor_' + str(id) + '.csv'
            else:
                fn = str(folder + '/') + str(filename) + '_' + str(id) + '.csv'

            csvfile = open(fn, 'w+')
            file_writer = csv.writer(csvfile, delimiter=',', lineterminator='\n')
            dict_file_writers[topicName] = file_writer
            dict_header_written[topicName] = False
            dict_csvfile_hdls[topicName] = csvfile

            if verbose:
                print("ROSbag_Pose2Ranges: creating csv file: %s " % fn)

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']

        found = True
        for id, topicName in dict_cfg["true_pose_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
                    pass
                pass
            if not found_:
                print("# ERROR: desired topic [" + str(topicName) + "] is not in bag file!")

            found = found & found_
            pass
        if not found:
            print("\nROSBag_Poses2RelPoses: not all topics found! Stopping here...")
            for topicName in dict_cfg["true_pose_topics"]:
                print(" * " + topicName)
            return False

        if verbose:
            print("\nROSBag_Poses2RelPoses: num messages " + str(num_messages))

        ## store the BODY SENSOR pose in a dictionary
        dict_T_BODY_SENSOR = dict() # map<topic, SE3>
        for ID, sensor_pos in dict_cfg["sensor_positions"].items():
            t_BT = np.array(sensor_pos)
            q_BT = UnitQuaternion()
            if ID in dict_cfg["sensor_orientations"].keys():
                # expecting: w,x,y,z
                q_BT = UnitQuaternion(np.array(dict_cfg["sensor_orientations"][ID]), norm=True)
                pass
            dict_T_BODY_SENSOR[dict_cfg["true_pose_topics"][ID]] = SE3.Rt(q_BT.R, t_BT, check=True)


        ## extract the poses of the desired topics from the BAG file
        round_decimals = 4
        dict_history = ROSBag_Pose.extract_poses(bag=bag, num_messages=num_messages,
                                                 dict_topic_pose_body=dict_cfg["true_pose_topics"],
                                                 dict_senor_topic_pose=dict_cfg["true_pose_topics"],
                                                 round_decimals=round_decimals,
                                                 dict_T_BODY_SENSOR=dict_T_BODY_SENSOR)

        ## TODO:
        try:  # else already exists
            for topic, hist in dict_history.items():
                if verbose:
                    print("\nROSBag_Poses2RelPoses: compute relative pose with respect to topic=[" + topic + "]...")
                for idx in tqdm(range(0, len(hist.t_vec)), total= len(hist.t_vec), unit="poses"):
                    file_writer = dict_file_writers[topic]
                    ID1 = get_key_from_value(dict_cfg["true_pose_topics"], topic)
                    T_GLOBAL_SENSOR1 = hist.val_vec[idx]
                    timestamp = hist.t_vec[idx]
                    for ID2, topic2 in dict_cfg["true_pose_topics"].items():
                        if topic != topic2:
                            #ID2 = get_key_from_value(dict_cfg["true_pose_topics"], topic2)

                            T_GLOBAL_SENSOR2 = None
                            if  dict_history[topic2].exists_at_t(timestamp) is None:
                                [t1, T_GLOBAL_BODY_T1] = dict_history[topic2].get_before_t(timestamp)
                                [t2, T_GLOBAL_BODY_T2] = dict_history[topic2].get_after_t(timestamp)

                                if t1 is None or t2 is None:
                                    #if verbose:
                                    #    print("* skip measurement from topic=" + topic + " at t=" + str(timestamp))
                                    continue
                                dt = t2 - t1
                                dt_i = timestamp - t1
                                i = abs(dt_i / dt)

                                # interpolate between poses:
                                q0 = UnitQuaternion(T_GLOBAL_BODY_T1.R)
                                q1 = UnitQuaternion(T_GLOBAL_BODY_T2.R)
                                p0 = T_GLOBAL_BODY_T1.t
                                p1 = T_GLOBAL_BODY_T2.t

                                qr = UnitQuaternion(qslerp(q0.vec, q1.vec, i, shortest=True), norm=True)
                                pr = p0 * (1 - i) + i * p1


                                T_GLOBAL_SENSOR2 = SE3.Rt(qr.R, pr, check=True)*dict_T_BODY_SENSOR[topic2]
                                if not SE3.isvalid(T_GLOBAL_SENSOR2, check=True):
                                    if T_GLOBAL_SENSOR2.A is None:
                                        if verbose:
                                            print(
                                                "* interp failed: skip measurement from topic=" + topic + " at t=" + str(
                                                    timestamp))
                                        continue
                                    else:
                                        q = UnitQuaternion(SO3(T_GLOBAL_SENSOR2.R, check=False), norm=True).unit()
                                        T_GLOBAL_SENSOR2 = SE3.Rt(q.R, T_GLOBAL_SENSOR2.t, check=True)
                                pass
                            else:
                                T_GLOBAL_SENSOR2 = dict_history[topic2].get_at_t(timestamp)*dict_T_BODY_SENSOR[topic2]

                            if not dict_header_written[topic]:
                                file_writer.writerow(["t", "ID1", "ID2", "tx", "ty", "tz", "qw", "qx", "qy", "qz", "range", "angle"])
                                dict_header_written[topic] = True

                            T_SENSOR1_SENSOR2 = SE3(T_GLOBAL_SENSOR1.inv() * T_GLOBAL_SENSOR2)
                            content = relpose_to_csv_line(timestamp, ID1, ID2, T_SENSOR1_SENSOR2, round_decimals)
                            file_writer.writerow(content)
        except  Exception as e:
            print("ROSBag_Poses2RelPoses: Unexpected error while creating the CSV files! msg=%s" % repr(e))
            print(str(sys.exc_info()))
            return False
        ## CLEANUP:
        # close all csv files
        for key, hdls in dict_csvfile_hdls.items():
            hdls.close()

        # check if a topic was found by checking if the topic header was written
        for topic, header_writen in dict_header_written.items():
            if not header_writen:
                print("\nROSBag_Poses2RelPoses: \n\tWARNING topic [" + str(topic) + "] was not in bag-file")
                print("\tbag file [" + str(bagfile_name) + "] contains: ")
                # print(info_dict['topics'])
                for t in info_dict['topics']:
                    print(t['topic'])
                return False

        if verbose:
            print("\nROSBag_Poses2RelPoses: extracting done! ")

        bag.close()
        return True


def main():
    # example: ROSBag_Pose2Ranges.py --bagfile ../test/sample_data//uwb_calib_a01_2023-08-31-21-05-46.bag --topic /d01/mavros/vision_pose/pose --cfg ../test/sample_data/config.yaml --verbose
    parser = argparse.ArgumentParser(
        description='ROSBag_Poses2RelPoses: extract given pose topics and computes for each pose a relative poses, which is stored into a CSV file')
    parser.add_argument('--bagfile', help='input bag file', required=True)
    parser.add_argument('--cfg', help='YAML configuration file describing the setup: {sensor_positions, sensor_topics}', default="config.yaml", required=True)
    parser.add_argument('--filename', help='csv filename of corresponding topic', default="")
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--verbose', action='store_true', default=False)

    tp_start = time.time()
    args = parser.parse_args()

    if ROSBag_Poses2RelPoses.extract(bagfile_name=args.bagfile, topic=str(args.topic), cfg=args.cfg,
                                     filename=args.filename, result_dir=args.result_dir,
                                     verbose=args.verbose
                                     ):
        print(" ")
        print("finished after [%s sec]\n" % str(time.time() - tp_start))
    else:
        print("failed! after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2020, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import math
import os

from cnspy_relative_pose_evaluation.ROSBag_Pose2AbsPoses import ROSBag_Pose2AbsPoses
from cnspy_relative_pose_evaluation.ROSBag_Poses2RelPoses import ROSBag_Poses2RelPoses
from cnspy_relative_pose_evaluation.ROSBag_TrueRelPoses import ROSBag_TrueRelPoses
from cnspy_relative_pose_evaluation.RelPose_ROSBag2CSV import RelPose_ROSBag2CSV
from cnspy_relative_pose_evaluation.RelPoseMeasEvaluationTool  import *
from cnspy_spatial_csv_formats.CSVSpatialFormatType import CSVSpatialFormatType
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory_evaluation.TrajectoryAlignmentTypes import TrajectoryAlignmentTypes
from cnspy_trajectory.Trajectory import Trajectory


from cnspy_trajectory_evaluation.TrajectoryEvaluationTool import TrajectoryEvaluationTool

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')
RES_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_results')



class EvaluationTrail:

    @staticmethod
    def test_ROSBag_Poses2RelPose():
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        ROSBag_Poses2RelPoses.extract(bagfile_name=bagfile_in,
                                      cfg=cfg_file,
                                      verbose=True,
                                      filename=None,
                                      result_dir=RES_DATA_DIR)

    @staticmethod
    def test_ROSBag_TrueRelPoses():
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        bagfile_out = str(RES_DATA_DIR + '/' + sequence + '_gt.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                    bagfile_out_name=bagfile_out,
                                     cfg=cfg_file,
                                     verbose=True)
    @staticmethod
    def test_ROSBag_TrueRelPoses_noise():
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        bagfile_out = str(RES_DATA_DIR + '/' + sequence + '_noisy.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                    bagfile_out_name=bagfile_out,
                                    cfg=cfg_file,
                                    stddev_pos=0.2,
                                    stddev_or=0.2,
                                    verbose=True)

    @staticmethod
    def test_RelPose_ROSbag2CSV():
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        RelPose_ROSBag2CSV.extract(bagfile_name=bagfile_in,
                                     cfg=cfg_file,
                                     verbose=True)

        fn_out = str(RES_DATA_DIR + '/all_meas_relposes.csv')
        RelPose_ROSBag2CSV.extract_to_one(bagfile_name=bagfile_in,
                                   cfg=cfg_file,
                                   fn=fn_out,
                                   result_dir=RES_DATA_DIR,
                                   verbose=True)

    @staticmethod
    def test_RelPoseEval_same():

        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')

        fn_out = str(RES_DATA_DIR + '/all_meas_relposes.csv')
        RelPose_ROSBag2CSV.extract_to_one(bagfile_name=bagfile_in,
                                   cfg=cfg_file,
                                   fn=fn_out,
                                   result_dir=RES_DATA_DIR,
                                   with_cov=True,
                                   verbose=True)

        Sensor_ID_arr = [0,1,2]
        fn_gt_ranges = fn_out
        fn_meas_ranges = fn_out
        # 4) evaluate the ranges
        cfg = AssociateRelPoseCfg(ID1=None,
                                 ID2=None,
                                 relative_timestamps=False,
                                 max_difference=10**-4,
                                 subsample=0,
                                 verbose=True,
                                 remove_outliers=True,
                                 max_range=9,
                                 range_error_val=0,
                                 label_timestamp='t',
                                 label_ID1='ID1',
                                 label_ID2='ID2',
                                 label_range='range')
        cfg.max_range_error = 9

        eval = RelPoseMeasEvaluation(fn_gt=fn_gt_ranges,
                                     fn_est=fn_meas_ranges,
                                     ID1_arr=Sensor_ID_arr,
                                     ID2_arr=Sensor_ID_arr,
                                     cfg=cfg,
                                     result_dir=str(RES_DATA_DIR + "/" + sequence + '/eval_same/'),
                                     prefix='',
                                     save_plot=True,
                                     show_plot=False,
                                     save_statistics=True,
                                     plot_timestamps=True,
                                     plot_ranges=True,
                                     plot_angles=True,
                                     plot_ranges_sorted=True,
                                     plot_range_error=True,
                                     plot_angle_error=True,
                                     plot_range_histogram=True,
                                     plot_angle_histogram=True,
                                     filter_histogram=True,
                                     plot_pose_err=True,
                                     plot_pose=True,
                                     verbose=True
                                     )

    @staticmethod
    def test_RelPoseEval():
        #sequence = 'two_spiral_2to3m'
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')

        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        RelPoseMeasEvaluationTool.evaluate(bagfile_in=bagfile_in,
                                           cfg_fn=cfg_file,
                                           result_dir=RES_DATA_DIR + "/" + sequence,
                                           save_plot=True,
                                           show_plot=False,
                                           remove_outliers=True,
                                           filter_histogram=True,
                                           verbose=True,
                                           ID_arr=[0,1,2],
                                           max_range=8,
                                           max_angle=0,
                                           interp_type=TrajectoryInterpolationType.cubic)

    @staticmethod
    def get_pos_rpy_stddev_static_test():
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_fn = str(SAMPLE_DATA_DIR + '/config.yaml')
        result_dir = RES_DATA_DIR + "/" + sequence,

        dict_cfg = None
        with open(cfg_fn, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "true_pose_topics" not in dict_cfg:
                print("[true_pose_topics] does not exist in fn=" + cfg_fn)
                return False

        bag = rosbag.Bag(bagfile_in)


        # map<true_pose_topic,History<timestamp, SE3>>
        dict_bspline, dict_history = ROSBag_TrueRelPoses.load_dict_splines(bag, dict_cfg, TrajectoryInterpolationType.cubic, min_dt=0.05)
        for key, hist_pose in dict_history.items():
            traj = Trajectory(hist_pose=hist_pose)
            pos_stddev = traj.get_pos_stddev()
            rpy_stddev = traj.get_rpy_stddev()
            print("Topic %s: pos_stddev: %s, rpy_stdev: %s" %( str(key), str(pos_stddev), str(rpy_stddev)))




    @staticmethod
    def test_RelPoseEval_gazebo():
        sequence = "dataset_octagon_loiter2_looped_merged_V2"
        bagfile_in = str("/home/jungr/workspace/catkin_ws/test_uvdar/src/aau/scripts/simulation/dataset_octagon_loiter2_looped_merged_sorted_2024-07-24-16-41-47.bag")

        cfg_file = str(SAMPLE_DATA_DIR + '/config_gazebo.yaml')
        RelPoseMeasEvaluationTool.evaluate(bagfile_in=bagfile_in,
                                           cfg_fn=cfg_file,
                                           result_dir=RES_DATA_DIR + "/" + sequence,
                                           save_plot=True,
                                           show_plot=False,
                                           verbose=True,
                                           ID_arr=[1,2])

    @staticmethod
    def test_RelPoseEval_synthetic():
        sequence = 'static_test'
        bagfile_in = str(RES_DATA_DIR + '/' + sequence + '_noisy.bag')

        cfg_file = str(SAMPLE_DATA_DIR + '/config.yaml')
        RelPoseMeasEvaluationTool.evaluate(bagfile_in=bagfile_in,
                                           cfg_fn=cfg_file,
                                           result_dir=RES_DATA_DIR + "/" + sequence + "_noisy",
                                           save_plot=True,
                                           show_plot=False,
                                           filter_histogram=False,
                                           verbose=True,
                                           ID_arr=[0,1])

    @staticmethod
    def test_RPLIDAR():
        sequence = 'two_spiral_one_hovers'
        #bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        #bagfile_in = '/home/jungr/workspace/datasets/uvdar_dataset/sequences/static_test/static_test.bag'
        bagfile_in = '/home/jungr/workspace/datasets/uvdar_dataset/sequences/two_spiral_one_hovers/two_spiral_one_hovers.bag'
        cfg_file = str(SAMPLE_DATA_DIR + '/config-lidar.yaml')
        TrajectoryEvaluationTool.evaluate(bagfile_in=bagfile_in,
                                           cfg=cfg_file,
                                           result_dir=RES_DATA_DIR + "/" + sequence + "/Lidar/",
                                           alignment_type=TrajectoryAlignmentTypes.posyaw,
                                           est_err_type = EstimationErrorType.type1,
                                           num_aligned_samples=1,
                                           plot = True,
                                           save_plot=True,
                                           show_plot=False,
                                           max_difference=0.01,
                                           relative_timestamps=True,
                                          fmt=CSVSpatialFormatType.Pose2DStamped,
                                          IDs=[0,1,2],
                                           verbose=True)

    @staticmethod
    def test_RelPoseEval_synthetic_abspos():
        sequence = 'two_spiral_2to3m'
        sequence = 'static_test'
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        bagfile_out = str(RES_DATA_DIR + '/' + sequence + '_abspose.bag')
        cfg_file = str(SAMPLE_DATA_DIR + '/config-abspose.yaml')
        ROSBag_Pose2AbsPoses.extract(bagfile_in=bagfile_in,
                                      bagfile_out=bagfile_out,
                                      cfg=cfg_file,
                                      stddev_pos=0.1,
                                      stddev_or=0.1,
                                      use_header_timestamp=False,
                                      verbose=True)

    @staticmethod
    def test_RelPoseEval_UVDAR_dataset():
        DATASET_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/sequences/"
        RES_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/eval/"
        sequence = 'static_test'
        #sequence = 'two_spiral_2to3m'
        #sequence = 'two_spiral_one_hovers'
        #sequence = 'loiter_spiral_yaxis'

        bagfile_in = str(DATASET_DIR + '/' + sequence + '/' + sequence + '.bag')

        cfg_file = str("/home/jungr/workspace/datasets/uvdar_dataset/tools/config-relpose.yaml")
        RelPoseMeasEvaluationTool.evaluate(bagfile_in=bagfile_in,
                                           cfg_fn=cfg_file,
                                           result_dir=RES_DIR + "/" + sequence,
                                           save_plot=True,
                                           show_plot=False,
                                           extra_plots=True,
                                           remove_outliers=True,
                                           filter_histogram=True,
                                           verbose=True,
                                           ID_arr=[0,1,2],
                                           max_range=8,
                                           max_angle=0,
                                           min_dt=0.05,
                                           interp_type=TrajectoryInterpolationType.cubic,
                                           pose_error_type=EstimationErrorType.type5)



if __name__ == "__main__":
     #EvaluationTrail.test_ROSBag_Poses2RelPose()
     #EvaluationTrail.test_ROSBag_TrueRelPoses()
     #EvaluationTrail.test_RelPose_ROSbag2CSV()
     #EvaluationTrail.test_ROSBag_TrueRelPoses_noise()
     #EvaluationTrail.test_RelPoseEval_same()
     #EvaluationTrail.test_RPLIDAR()
     #EvaluationTrail.test_RelPoseEval_synthetic_abspos()
     #EvaluationTrail.get_pos_rpy_stddev_static_test()
     EvaluationTrail.test_RelPoseEval_UVDAR_dataset()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import sys
import rosbag
import time
import os
import argparse
import yaml
from tqdm import tqdm
from cnspy_trajectory.HistoryBuffer import get_key_from_value

from mrs_msgs.msg import PoseWithCovarianceArrayStamped, PoseWithCovarianceIdentified




class ROSBag_ModifyRelPoses:
    def __init__(self):
        pass

    @staticmethod
    def extract(bagfile_in_name,
                bagfile_out_name,
                cfg,
                use_header_timestamp=False,
                verbose=False,
                use_position=True,
                use_orientation=True,
                ):

        if not os.path.isfile(bagfile_in_name):
            print("ROSBag_ModifyRelPoses: could not find file: %s" % bagfile_in_name)
            return False
        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("ROSBag_ModifyRelPoses: could not find file: %s" % cfg)
            return False

        if verbose:
            print("ROSBag_ModifyRelPoses:")
            print("* bagfile in name: " + str(bagfile_in_name))
            print("* bagfile out name: " + str(bagfile_out_name))
            print("* cfg YAML file: \t " + str(cfg))
            print("* use_position: " + str(use_position))
            print("* use_orientation: " + str(use_orientation))
            print("* use_header_timestamp: " + str(use_header_timestamp))

        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_in_name)
        except:
            if verbose:
                print("ROSBag_ModifyRelPoses: Unexpected error!")

            return False

        ## create result dir:
        [root, ext] = os.path.splitext(bagfile_out_name)
        [head, tail] = os.path.split(root)
        try:  # else already exists
            os.makedirs(head)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(head))

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "relpose_topics" not in dict_cfg:
                print("[relpose_topics] does not exist in fn=" + cfg)
                return False
            print("Read successful")
        if verbose:
            print("configuration contains:")
            print("relpose_topics:" + str(dict_cfg["relpose_topics"]))

        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("ROSBag_ModifyRelPoses: Unexpected error, bag file might be empty!")
            bag.close()
            return False

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']

        # check if desired topics are in
        for id, topicName in dict_cfg["relpose_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
            if not found_:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")

        ## TODO:
        cnt = 0
        try:  # else already exists
            print("ROSBag_ModifyRelPoses: modifying relative pose measurements...")
            with rosbag.Bag(bagfile_out_name, 'w') as outbag:
                for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
                    if topic in dict_cfg["relpose_topics"].values() and hasattr(msg, 'poses') and hasattr(msg, 'header'):
                        ID1 = get_key_from_value(dict_cfg["relpose_topics"], topic)
                        idx_pose = 0
                        timestamp = msg.header.stamp.to_sec()

                        msg_gt = PoseWithCovarianceArrayStamped()
                        msg_gt.header = msg.header

                        for relpose in msg.poses: # id, pose, covariance
                            ID2 = relpose.id
                            pos_id = PoseWithCovarianceIdentified()
                            pos_id.id = ID2
                            # TODO: add noise etc.
                            if use_position:
                                pos_id.pose.position = relpose.pose.position
                            else:
                                pos_id.pose.position.x = 0
                                pos_id.pose.position.y = 0
                                pos_id.pose.position.z = 0

                            if use_orientation:
                                pos_id.pose.orientation = relpose.pose.orientation
                            else:
                                pos_id.pose.orientation.w = 1
                                pos_id.pose.orientation.x = 0
                                pos_id.pose.orientation.y = 0
                                pos_id.pose.orientation.z = 0

                            # copy covariance
                            if use_position and use_orientation:
                                pos_id.covariance = relpose.covariance
                            elif use_position:
                                # upper left block
                                for idx in [0,1,2,6,7,8,12,13,14]:
                                    pos_id.covariance[idx] = relpose.covariance[idx]
                            elif use_orientation:
                                # lower right block
                                for idx in [21,22,23,27,28,29,33,34,35]:
                                    pos_id.covariance[idx] = relpose.covariance[idx]

                            msg_gt.poses.append(pos_id)
                            idx_pose += 1
                        # for id2
                        if use_header_timestamp and hasattr(msg, "header"):
                            outbag.write(topic, msg_gt, msg.header.stamp)
                        else:
                            outbag.write(topic, msg_gt, t)
                        cnt += 1

                    else:
                        # write other topic
                        if use_header_timestamp and hasattr(msg, "header"):
                            outbag.write(topic, msg, msg.header.stamp)
                        else:
                            outbag.write(topic, msg, t)
                        pass




        except  Exception as e:
            print("ROSBag_ModifyRelPoses: Unexpected error while creating the bag file! msg=%s" % repr(e))
            print(str(sys.exc_info()))
            return False
        ## CLEANUP:
        except AssertionError as error:
            print(error)
            print("ROSBag_ModifyRelPoses: Unexpected error while creating bag file")
            return False

        if verbose:
            print("\nROSBag_ModifyRelPoses: " + str(cnt) + " range measurements modified!")
        bag.close()
        return True



def main():
    # example: ROSBag_Pose2Ranges.py --bagfile ../test/sample_data//uwb_calib_a01_2023-08-31-21-05-46.bag --topic /d01/mavros/vision_pose/pose --cfg ../test/sample_data/config.yaml --verbose
    parser = argparse.ArgumentParser(
        description='ROSBag_ModifyRelPoses: extract given relative pose topics and modifies them (removing position or orientation')
    parser.add_argument('--bagfile_in', help='input bag file', required=True)
    parser.add_argument('--bagfile_out', help='output bag file', required=True)
    parser.add_argument('--cfg',
                        help='YAML configuration file describing the setup: {rel_tag_positions, abs_anchor_positions}',
                        default="config.yaml", required=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--use_header_timestamp', action='store_true',
                        help='overwrites the bag time with the header time stamp', default=False)
    parser.add_argument('--remove_position', action='store_true',
                        help='clears the position of the UVDAR measurement', default=False)
    parser.add_argument('--remove_orientation', action='store_true',
                        help='clears the orientation of the UVDAR measurement', default=False)
    tp_start = time.time()
    args = parser.parse_args()

    if ROSBag_ModifyRelPoses.extract(bagfile_in_name=args.bagfile_in,
                                 bagfile_out_name=args.bagfile_out,
                                 cfg=args.cfg,
                                 verbose=args.verbose,
                                 use_header_timestamp=args.use_header_timestamp,
                                 use_position=not args.remove_position,
                                 use_orientation=not args.remove_orientation):
        print(" ")
        print("finished after [%s sec]\n" % str(time.time() - tp_start))
    else:
        print("failed! after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import unittest
import os

from cnspy_relative_pose_evaluation.ROSBag_TrueRelPoses import *
from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.PlotLineStyle import PlotLineStyle
from cnspy_trajectory.Trajectory import Trajectory
from cnspy_trajectory.TrajectoryPlotConfig import TrajectoryPlotConfig
from cnspy_trajectory.TrajectoryPlotter import TrajectoryPlotter
from cnspy_trajectory_evaluation.TrajectoryAlignmentTypes import TrajectoryAlignmentTypes
from cnspy_trajectory_evaluation.TrajectoryEvaluation import TrajectoryEvaluation

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')
RES_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_results')




class ROSBag_TrueRelPoses_Test(unittest.TestCase):


    def test_interpolation(self):
        sequence = "static_test"
        #sequence = 'two_spiral_2to3m'
        cfg_fn = str(SAMPLE_DATA_DIR + '/config.yaml')
        result_dir = RES_DATA_DIR + "/" + sequence
        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')

        bag = rosbag.Bag(bagfile_in)
        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)
        num_messages = num_messages = info_dict['messages']
        verbose = True
        ignore_new_topic_name = True
        interp_type = TrajectoryInterpolationType.cubic
        min_dt = 0.1
        dict_cfg = ROSBag_TrueRelPoses.load_dict_cfg(cfg_fn=cfg_fn,
                                                     ignore_new_topic_name=ignore_new_topic_name,
                                                     verbose=verbose)
        round_decimals = 4
        dict_bsplines, dict_history = ROSBag_TrueRelPoses.load_dict_splines(bag, dict_cfg, interp_type, min_dt,
                                                                              num_messages, round_decimals)

        for key, bspline in dict_bsplines.items():
            t_arr = dict_history[key].t_vec

            dict_pose = dict_history[key]
            if dict_pose and bspline and len(dict_pose.t_vec):
                traj_in = Trajectory(hist_pose=dict_pose)
                traj_out = bspline.get_trajectory(t_arr=traj_in.t_vec,interp_type=interp_type)

                cfg = TrajectoryPlotConfig(show=True,
                                           close_figure=False,
                                           save_fn=result_dir + '/bspline_test/' + str(interp_type) + "/" + key.replace("/", "-") + "_mindt_" + str(min_dt) + ".png")
                TrajectoryPlotter.multi_plot_3D([traj_in, traj_out],
                                                cfg=cfg,
                                                name_list=['gt-'+key, 'interp-'+key],
                                                ls=PlotLineStyle(linestyle='--'))


                TE = TrajectoryEvaluation(traj_gt=traj_out,
                                          traj_est=traj_in,
                                          result_dir=result_dir + '/bspline_test/' + str(interp_type) + "/" + key.replace("/", "-") + "_mindt_" + str(min_dt) + "/",
                                          alignment_type=TrajectoryAlignmentTypes.none,
                                          plot=True,
                                          est_err_type=EstimationErrorType.type5,
                                          verbose=True)


    def test_extract(self):
        #sequence = "static_test"
        sequence = 'two_spiral_2to3m'

        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_fn = str(SAMPLE_DATA_DIR + '/config.yaml')
        result_dir = RES_DATA_DIR + "/" + sequence

        bagfile_out = str(result_dir +  '/true-ranges.bag')
        res = ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                          bagfile_out_name=bagfile_out,
                                          cfg=cfg_fn,
                                          verbose=True,
                                          use_header_timestamp=False,
                                          ignore_new_topic_name=True,
                                          stddev_pos=0.0,
                                          interp_type=TrajectoryInterpolationType.cubic,
                                          min_dt=0.05)

        self.assertEqual(True, res)  # add as

        res = ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                          bagfile_out_name=bagfile_out,
                                          cfg=cfg_fn,
                                          verbose=True,
                                          use_header_timestamp=False,
                                          ignore_new_topic_name=True,
                                          stddev_pos=0.0,
                                          interp_type=TrajectoryInterpolationType.linear,
                                          min_dt=0.05)

        self.assertEqual(True, res)  # add as
        pass


    def test_add_noise(self):
        sequence = "static_test"
        #sequence = 'two_spiral_2to3m'

        bagfile_in = str(SAMPLE_DATA_DIR + '/' + sequence + '.bag')
        cfg_fn = str(SAMPLE_DATA_DIR + '/config.yaml')
        result_dir = RES_DATA_DIR + "/" + sequence + "/"

        bagfile_out = str(result_dir +  '/noisy-ranges-type1.bag')
        res = ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                          bagfile_out_name=bagfile_out,
                                          cfg=cfg_fn,
                                          verbose=True,
                                          use_header_timestamp=False,
                                          ignore_new_topic_name=True,
                                          stddev_pos=0.1,
                                          stddev_or=0.1,
                                          pose_error_type=EstimationErrorType.type1,
                                          interp_type=TrajectoryInterpolationType.cubic,
                                          min_dt=0.05)

        bagfile_out = str(result_dir +  '/noisy-ranges-type5.bag')
        res = ROSBag_TrueRelPoses.extract(bagfile_in_name=bagfile_in,
                                          bagfile_out_name=bagfile_out,
                                          cfg=cfg_fn,
                                          verbose=True,
                                          use_header_timestamp=False,
                                          ignore_new_topic_name=True,
                                          stddev_pos=0.1,
                                          stddev_or=0.1,
                                          pose_error_type=EstimationErrorType.type5,
                                          interp_type=TrajectoryInterpolationType.cubic,
                                          min_dt=0.05)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2024, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################################################################
import unittest
import os

from cnspy_relative_pose_evaluation.RelPoseMeasEvaluation import *

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sample_data')
RES_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_results')


class RelPoseMeasEvaluation_Test(unittest.TestCase):
    def test_plot_NEES(self):
        dict_fn_gt = dict({'GT': SAMPLE_DATA_DIR + '/synth/meas_B1B2_N1000_ref_np-0_nR-0.csv',
                           'CUBIC': SAMPLE_DATA_DIR + '/synth/meas_B1B2_N970_SUB10_CUBIC_np-0.01_nR-0.01.csv',
                           'LINEAR': SAMPLE_DATA_DIR + '/synth/meas_B1B2_N991_SUB10_LINEAR_np-0.01_nR-0.01.csv'
                           })
        for key,fn_gt in dict_fn_gt.items():
            err_types = list([EstimationErrorType.type1, EstimationErrorType.type2, EstimationErrorType.type5])
            #err_types = list([EstimationErrorType.type1])
            for err_type in err_types:
                fn_est = SAMPLE_DATA_DIR + '/synth/meas_B1B2_N1000_'+str(err_type)+'_np-0.1_nR-0.1.csv'

                RelPoseMeasEvaluation(fn_gt, fn_est,
                                      result_dir=RES_DATA_DIR + '/RPME/REF_'+ str(key) + '/' + str(err_type) + '/',
                                      cfg=AssociateRelPoseCfg(pose_error_type=err_type),
                                      save_plot=True,
                                      show_plot=False,
                                      filter_histogram=False,
                                      plot_timestamps=False,
                                      plot_ranges=False,
                                      plot_angles=False,
                                      plot_ranges_sorted=False,
                                      plot_range_error=False,
                                      plot_angle_error=False,
                                      plot_range_histogram=True,
                                      plot_angle_histogram=True,
                                      plot_position_err=False,
                                      plot_pose_err=False,
                                      plot_pose=False,
                                      plot_NEES=False,
                                      verbose=True,
                                      save_statistics=True)


    def test_with_noise(self):
        fn_gt = SAMPLE_DATA_DIR + '/synth/meas_B1B2_N1000_ref_np-0_nR-0.csv'

        err_types = list([EstimationErrorType.type1, EstimationErrorType.type2, EstimationErrorType.type5])
        # err_types = list([EstimationErrorType.type1])
        for err_type in err_types:
            fn_est = SAMPLE_DATA_DIR + '/synth/meas_B1B2_N1000_'+str(err_type)+'_np-0.1_nR-0.1.csv'

            RelPoseMeasEvaluation(fn_gt, fn_est,
                                  result_dir=RES_DATA_DIR + '/RPME/REF_GT/'+ str(err_type) + '/',
                                  cfg=AssociateRelPoseCfg(pose_error_type=err_type),
                                  save_plot=True,
                                  show_plot=False,
                                  filter_histogram=False,
                                  plot_timestamps=False,
                                  plot_ranges=True,
                                  plot_angles=True,
                                  plot_ranges_sorted=True,
                                  plot_range_error=True,
                                  plot_angle_error=True,
                                  plot_range_histogram=True,
                                  plot_angle_histogram=True,
                                  plot_position_err=True,
                                  plot_pose_err=True,
                                  plot_pose=True,
                                  plot_NEES=True,
                                  verbose=True,
                                  save_statistics=True)

        self.assertEqual(True, True)  # add assertion here

    def test_dataset_eval(self):
        DATASET_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/eval/"
        RES_DIR= "/home/jungr/workspace/datasets/uvdar_dataset/eval/"

        sequence = 'static_test'
        #sequence = 'two_spiral_2to3m'
        #sequence = 'two_spiral_one_hovers'
        #sequence = 'loiter_spiral_yaxis'

        fn_gt = DATASET_DIR + '/sequences/' + sequence +  '/all-true-ranges.csv'
        fn_est = DATASET_DIR + '/sequences/' + sequence +  '/all-meas-ranges.csv'
        RelPoseMeasEvaluation(fn_gt, fn_est,
                              result_dir=RES_DIR + sequence + '/',
                              cfg=AssociateRelPoseCfg(pose_error_type=EstimationErrorType.type5, remove_outliers=True,  max_range=8),
                              save_plot=True,
                              show_plot=True,
                              filter_histogram=False,
                              plot_timestamps=False,
                              plot_ranges=False,
                              plot_angles=False,
                              ID1_arr=[2],
                              ID2_arr=[1],
                              plot_ranges_sorted=False,
                              plot_range_error=False,
                              plot_angle_error=False,
                              plot_range_histogram=False,
                              plot_angle_histogram=False,
                              plot_position_err=True,
                              plot_pose_err=False,
                              plot_pose=False,
                              plot_NEES=False,
                              verbose=True,
                              save_statistics=False)



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cnspy_relative_pose_evaluation',
    version="0.1.16",
    author='Roland Jung',
    author_email='roland.jung@aau.at',
    description='Evaluation of uncertain relative pose measurements.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/aau-cns/cnspy_relative_pose_evaluation/',
    project_urls={
        "Bug Tracker": "https://github.com/aau-cns/cnspy_relative_pose_evaluation/issues",
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],

    packages=find_packages(exclude=["test_*", "TODO*"]),
    python_requires='>=3.6',
    install_requires=['numpy',
                      'pandas',
                      'spatialmath-python',
                      'scipy',
                      'matplotlib',
                      'joblib',
                      'configparser',
                      'cnspy_numpy_utils',
                      'cnspy_timestamp_association>=0.2.1',
                      'cnspy_csv2dataframe>=0.2.3',
                      'cnspy_trajectory>=0.2.11',
                      'cnspy_trajectory_evaluation>=0.2.7',
                      'cnspy_ranging_evaluation>=0.2.9'],
    entry_points={
        'console_scripts': [
            'RelPoseMeasEvaluationTool = cnspy_relative_pose_evaluation.RelPoseMeasEvaluationTool:main',
            'RelPoseMeasEvaluation = cnspy_relative_pose_evaluation.RelPoseMeasEvaluation:main',
            'RelPose_ROSBag2CSV = cnspy_relative_pose_evaluation.RelPose_ROSBag2CSV:main',
            'ROSBag_TrueRelPoses = cnspy_relative_pose_evaluation.ROSBag_TrueRelPoses:main',
            'ROSBag_Poses2RelPoses = cnspy_relative_pose_evaluation.ROSBag_Poses2RelPoses:main',
            'ROSBag_ModifyRelPoses = cnspy_relative_pose_evaluation.ROSBag_ModifyRelPoses:main',
            'ROSBag_Pose2AbsPoses = cnspy_relative_pose_evaluation.ROSBag_Pose2AbsPoses:main',
        ],
    },
)






#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################
import math
import sys

import rosbag
import time
import os
import argparse
import yaml
from tqdm import tqdm
import numpy as np
from spatialmath import UnitQuaternion, SO3, SE3
from spatialmath.base.quaternions import qslerp

from cnspy_spatial_csv_formats.EstimationErrorType import EstimationErrorType
from cnspy_trajectory.BsplineSE3 import BsplineSE3, TrajectoryInterpolationType
from cnspy_trajectory.HistoryBuffer import get_key_from_value
from cnspy_trajectory.ROSBag_Pose import ROSBag_Pose
from cnspy_trajectory.SpatialConverter import SpatialConverter

from mrs_msgs.msg import PoseWithCovarianceArrayStamped, PoseWithCovarianceIdentified


def interpolate_pose(pose_hist, timestamp, round_decimals=4) -> SE3:
    if pose_hist is None:
        return None

    timestamp = round(timestamp, round_decimals)
    [ti, T_GLOBAL_BODY_Ti] = pose_hist.get_at_t(timestamp)
    if T_GLOBAL_BODY_Ti is None:
        [t1, T_GLOBAL_BODY_T1] = pose_hist.get_before_t(timestamp)
        [t2, T_GLOBAL_BODY_T2] = pose_hist.get_after_t(timestamp)
        if t1 is None or t2 is None:
            # if verbose:
            #    print("* skip measurement from topic=" + topic + " at t=" + str(timestamp))
            return None

        assert (t1 <= t2), "Timestamp not ascending!"

        dt = t2 - t1
        dt_i = timestamp - t1
        i = abs(dt_i / dt)

        if dt < 10 ** -3:
            return T_GLOBAL_BODY_T1
        if i < 10 ** -2:
            return T_GLOBAL_BODY_T1
        elif i > 1.0 - 10 ** -2:
            return T_GLOBAL_BODY_T2
        # if dt > 1.0:
        #    print("WARNING: interpolate_pose(): dt="+str(dt) + " is huge!")

        # interpolate between poses:
        q0 = UnitQuaternion(T_GLOBAL_BODY_T1.R)
        q1 = UnitQuaternion(T_GLOBAL_BODY_T2.R)
        p0 = T_GLOBAL_BODY_T1.t
        p1 = T_GLOBAL_BODY_T2.t

        qi = UnitQuaternion(qslerp(q0.vec, q1.vec, i, shortest=True), norm=True)
        pi = p0 * (1 - i) + i * p1

        T_GLOBAL_BODY_Ti = SE3.Rt(qi.R, pi, check=True)
        if not SE3.isvalid(T_GLOBAL_BODY_Ti, check=True):
            if T_GLOBAL_BODY_Ti.A is None:
                return None
            else:
                q = UnitQuaternion(SO3(T_GLOBAL_BODY_Ti.R, check=False), norm=True).unit()
                T_GLOBAL_BODY_Ti = SE3.Rt(q.R, T_GLOBAL_BODY_Ti.t, check=True)
    return T_GLOBAL_BODY_Ti


def random_orientation(stddev_rad):
    ax = np.random.randn(3, 1)
    ax = ax / np.linalg.norm(ax)
    ang = np.random.randn(1, 1)[0] * stddev_rad
    return axang2quat(ax, ang[0])


def axang2quat(u, phi):
    u = u / np.linalg.norm(u)
    q_w = math.cos(phi * 0.5)
    q_v = u * math.sin(phi * 0.5)
    return UnitQuaternion(s=q_w, v=np.array(q_v)).unit()


class ROSBag_TrueRelPoses:
    def __init__(self):
        pass

    @staticmethod
    def load_dict_cfg(cfg_fn, ignore_new_topic_name, verbose) -> dict:
        with open(cfg_fn, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "sensor_positions" not in dict_cfg:
                print("[sensor_positions] does not exist in fn=" + cfg_fn)
                return None
            if "sensor_orientations" not in dict_cfg:
                print("[sensor_orientations] does not exist in fn=" + cfg_fn)
            if "object_positions" not in dict_cfg:
                print("[object_positions] does not exist in fn=" + cfg_fn)
                # check if object_positions ids are different from sensor_positions
                for key, val in dict_cfg["object_postions"].items():
                    if key in dict_cfg["sensor_positions"].keys():
                        print("[object_positions] id " + str(key) + " is not unique!")
                        return None
            if "object_orientations" not in dict_cfg:
                print("[object_orientations] does not exist in fn=" + cfg_fn)
                for key, val in dict_cfg["object_orientations"].items():
                    if key in dict_cfg["sensor_positions"].keys():
                        print("[object_orientations] id " + str(key) + " is not unique!")
                        return None
            if "true_pose_topics" not in dict_cfg:
                print("[true_pose_topics] does not exist in fn=" + cfg_fn)
                return None
            if "relpose_topics" not in dict_cfg:
                print("[relpose_topics] does not exist in fn=" + cfg_fn)
                return None
            if "new_relpose_topics" not in dict_cfg:
                print("[new_relpose_topics] does not exist in fn=" + cfg_fn)
            elif ignore_new_topic_name:
                print("[new_relpose_topics] WILL BE IGNORED, but was specified in fn=" + cfg_fn)

            for key, val in dict_cfg["true_pose_topics"].items():
                if key not in dict_cfg["relpose_topics"].keys() or key not in dict_cfg["sensor_positions"].keys():
                    print("[true_pose_topics] id " + str(key) + " is not present in relpose_topics or sensor_positions!")
                    return None

            print("Configuration read successfully")
        if verbose:
            print("configuration contains:")
            print("sensor_positions:" + str(dict_cfg["sensor_positions"]))
            print("sensor_orientations:" + str(dict_cfg["sensor_orientations"]))
            print("object_positions:" + str(dict_cfg["object_positions"]))
            print("object_orientations:" + str(dict_cfg["object_orientations"]))
            print("true_pose_topics:" + str(dict_cfg["true_pose_topics"]))
            print("relpose_topics:" + str(dict_cfg["relpose_topics"]))
            if "new_relpose_topics" in dict_cfg and not ignore_new_topic_name:
                print("new_relpose_topics:" + str(dict_cfg["new_relpose_topics"]))

        return dict_cfg

    @staticmethod
    def add_noise(p, R, stddev_pos:float, stddev_or:float, pose_error_type:EstimationErrorType):
        if stddev_pos == 0 and stddev_or == 0:
            return p, R

        if stddev_pos > 0:
            n_p = np.random.normal(0, stddev_pos, size=3)
        else:
            n_p = np.zeros(3)

        if stddev_or > 0:
            n_R = np.random.normal(0, stddev_or, size=3)
        else:
            n_R = np.zeros(3)

        if pose_error_type == EstimationErrorType.type5:
            if stddev_pos > 0:
                p = p + n_p.T
            if stddev_or > 0:
                q_noise = random_orientation(stddev_rad=stddev_or)
                R = np.matmul(R, q_noise.R)
            return p, R

        elif pose_error_type == EstimationErrorType.type1 or pose_error_type == EstimationErrorType.type2:
            n_T = np.zeros(6)
            n_T[0:3] = n_p
            n_T[3:6] = n_R
            T_n = SpatialConverter.theta_to_T(n_T)
            T = np.identity(4)
            T[0:3, 0:3] = R
            T[:3, 3] = p
            if pose_error_type == EstimationErrorType.type1:
                T_AB = SE3(np.matmul(T, T_n), check=False)
                return T_AB.t, T_AB.R
            else:
                T_AB = SE3(np.matmul(T_n, T), check=False)
                return T_AB.t, T_AB.R
        else:
            assert False, "EstimationErrorType not supported!"
            return p, R

    @staticmethod
    def extract(bagfile_in_name,
                bagfile_out_name,
                cfg,
                stddev_pos=0.0,
                stddev_or=0.0,
                use_header_timestamp=False,
                verbose=False,
                replace_with_new_topic=False,
                ignore_new_topic_name=False,
                interp_type=TrajectoryInterpolationType.cubic,
                min_dt=0.05,
                pose_error_type: EstimationErrorType = EstimationErrorType.type5
                ):

        if not os.path.isfile(bagfile_in_name):
            print("ROSBag_TrueRelPoses: could not find file: %s" % bagfile_in_name)
            return False
        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("ROSBag_TrueRelPoses: could not find file: %s" % cfg)
            return False

        if verbose:
            print("ROSBag_TrueRelPoses:")
            print("* bagfile in name: " + str(bagfile_in_name))
            print("* bagfile out name: " + str(bagfile_out_name))
            print("* cfg YAML file: \t " + str(cfg))
            print("* stddev_pos: " + str(stddev_pos))
            print("* stddev_or: " + str(stddev_or))
            print("* use_header_timestamp: " + str(use_header_timestamp))
            print("* replace_with_new_topic: " + str(replace_with_new_topic))

        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_in_name)
        except:
            if verbose:
                print("ROSBag_TrueRelPoses: Unexpected error!")

            return False

        ## create result dir:
        [root, ext] = os.path.splitext(bagfile_out_name)
        [head, tail] = os.path.split(root)
        try:  # else already exists
            os.makedirs(head)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(head))

        dict_cfg = ROSBag_TrueRelPoses.load_dict_cfg(cfg_fn=cfg,
                                                     ignore_new_topic_name=ignore_new_topic_name,
                                                     verbose=verbose)
        if dict_cfg is None:
            print(" ERROR configuration is wrong!")
            return False

        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("ROSBag_TrueRelPoses: Unexpected error, bag file might be empty!")
            bag.close()
            return False

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']

        found = ROSBag_TrueRelPoses.check_topics(bag_topics, dict_cfg, num_messages, verbose)
        if verbose and not found:
            print("ROSBag_TrueRelPoses: desired topics not found!")

        dict_static_objects = ROSBag_TrueRelPoses.load_dict_static_objects(dict_cfg)
        round_decimals = 4
        dict_bsplines, dict_history = ROSBag_TrueRelPoses.load_dict_splines(bag, dict_cfg, interp_type, min_dt,
                                                                            num_messages, round_decimals)


        if len(dict_bsplines) == 0:
            if verbose:
                print("ROSBag_TrueRelPoses: No poses found!")
            bag.close()
            return False

        cnt = 0
        try:  # else already exists
            print("ROSBag_TrueRelPoses: computing new relative pose measurements...")
            with rosbag.Bag(bagfile_out_name, 'w') as outbag:
                for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
                    if topic in dict_cfg["relpose_topics"].values() and hasattr(msg, 'poses') and hasattr(msg,
                                                                                                          'header'):
                        ID1 = get_key_from_value(dict_cfg["relpose_topics"], topic)
                        idx_pose = 0
                        timestamp = msg.header.stamp.to_sec()

                        msg_gt = PoseWithCovarianceArrayStamped()
                        msg_gt.header = msg.header

                        for relpose in msg.poses:  # id, pose, covariance
                            ID2 = relpose.id

                            is_sensor = (ID2 in dict_cfg["true_pose_topics"].keys())
                            is_object = (ID2 in dict_cfg["object_positions"].keys())
                            if is_sensor or is_object:
                                # ID2 = get_key_from_value(dict_cfg["true_pose_topics"], topic2)

                                T_GLOBAL_SENSOR1 = dict_bsplines[dict_cfg["true_pose_topics"][ID1]].get_pose(
                                    t=timestamp,
                                    interp_type=interp_type,
                                    round_decimals=round_decimals)

                                if is_sensor:
                                    T_GLOBAL_SENSOR2 = dict_bsplines[dict_cfg["true_pose_topics"][ID2]].get_pose(
                                        t=timestamp,
                                        interp_type=interp_type,
                                        round_decimals=round_decimals)
                                else:
                                    T_GLOBAL_SENSOR2 = dict_static_objects[ID2]

                                if T_GLOBAL_SENSOR1 is not None and T_GLOBAL_SENSOR2 is not None:
                                    T_SENSOR1_SENSOR2 = T_GLOBAL_SENSOR1.inv() * T_GLOBAL_SENSOR2

                                    t_, R_ = ROSBag_TrueRelPoses.add_noise(p=T_SENSOR1_SENSOR2.t,
                                                                           R=T_SENSOR1_SENSOR2.R,
                                                                           stddev_pos=stddev_pos,
                                                                           stddev_or=stddev_or,
                                                                           pose_error_type=pose_error_type)
                                    p = t_
                                    q = UnitQuaternion(R_, norm=True).unit()
                                    qv = q.vec

                                    pos_id = PoseWithCovarianceIdentified()
                                    pos_id.id = ID2
                                    pos_id.pose.position.x = p[0]
                                    pos_id.pose.position.y = p[1]
                                    pos_id.pose.position.z = p[2]
                                    pos_id.pose.orientation.w = qv[0]
                                    pos_id.pose.orientation.x = qv[1]
                                    pos_id.pose.orientation.y = qv[2]
                                    pos_id.pose.orientation.z = qv[3]

                                    if stddev_pos > 0:
                                        # upper left block
                                        var_pos = stddev_pos * stddev_pos
                                        for idx in [0, 7, 14]:
                                            pos_id.covariance[idx] = var_pos
                                    if stddev_or > 0:
                                        # lower right block
                                        var_or = stddev_or * stddev_or
                                        for idx in [21, 28, 35]:
                                            pos_id.covariance[idx] = var_or
                                    msg_gt.poses.append(pos_id)
                            idx_pose += 1
                        # for id2
                        topic_out = topic
                        if "new_relpose_topics" in dict_cfg and not ignore_new_topic_name:
                            id_topic = ID1
                            if id_topic in dict_cfg["new_relpose_topics"]:
                                # assign new topic name
                                topic_out = dict_cfg["new_relpose_topics"][id_topic]
                                if not replace_with_new_topic:
                                    # store original message, otherwise it gets lost:
                                    if use_header_timestamp and hasattr(msg, "header"):
                                        outbag.write(topic, msg, msg.header.stamp)
                                    else:
                                        outbag.write(topic, msg, t)
                            else:
                                print("No new_relpose_topics with ID=[" + str(id_topic) + "] found!")
                                return False

                        if use_header_timestamp and hasattr(msg, "header"):
                            outbag.write(topic_out, msg_gt, msg.header.stamp)
                        else:
                            outbag.write(topic_out, msg_gt, t)
                        cnt += 1

                    else:
                        # write other topic
                        if use_header_timestamp and hasattr(msg, "header"):
                            outbag.write(topic, msg, msg.header.stamp)
                        else:
                            outbag.write(topic, msg, t)
                        pass




        except  Exception as e:
            print("ROSBag_TrueRelPoses: Unexpected error while creating the bag file! msg=%s" % repr(e))
            print(str(sys.exc_info()))
            return False
        ## CLEANUP:
        except AssertionError as error:
            print(error)
            print("ROSBag_TrueRelPoses: Unexpected error while creating bag file")
            return False

        if verbose:
            print("\nROSBag_TrueRelPoses: " + str(cnt) + " range measurements modified!")
        bag.close()
        return True

    @staticmethod
    def check_topics(bag_topics, dict_cfg, num_messages, verbose) -> bool:
        # check if desired topics are in
        found = True
        for ID, topicName in dict_cfg["true_pose_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
            if not found_:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")
                found = False
        for ID, topicName in dict_cfg["relpose_topics"].items():
            found_ = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found_ = True
            if not found_:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")
                found = False
        if verbose:
            print("\nROSBag_TrueRelPoses: num messages " + str(num_messages))
        return found

    @staticmethod
    def load_dict_static_objects(dict_cfg):
        dict_T_GLOBAL_OBJECT = dict()  # map<topic, SE3>
        for ID, sensor_pos in dict_cfg["object_positions"].items():
            t_BT = np.array(sensor_pos)
            q_BT = UnitQuaternion()
            if ID in dict_cfg["object_orientations"].keys():
                # expecting: w,x,y,z
                q_vec = np.array(dict_cfg["object_orientations"][ID])
                q_BT = UnitQuaternion(q_vec).unit()
                pass
            dict_T_GLOBAL_OBJECT[ID] = SE3.Rt(q_BT.R, t_BT, check=True)

        return dict_T_GLOBAL_OBJECT

    @staticmethod
    def load_dict_splines(bag, dict_cfg, interp_type, min_dt=0.05, num_messages=None, round_decimals=4):
        if num_messages is None:
            info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)
            num_messages = info_dict['messages']

        ## store the BODY SENSOR pose in a dictionary
        dict_T_BODY_SENSOR = dict()  # map<topic, SE3>
        for ID, sensor_pos in dict_cfg["sensor_positions"].items():
            t_BT = np.array(sensor_pos)
            q_BT = UnitQuaternion()
            if ID in dict_cfg["sensor_orientations"].keys():
                # expecting: w,x,y,z
                q_vec = np.array(dict_cfg["sensor_orientations"][ID])
                q_BT = UnitQuaternion(q_vec).unit()
                pass
            dict_T_BODY_SENSOR[dict_cfg["true_pose_topics"][ID]] = SE3.Rt(q_BT.R, t_BT, check=True)
        ## extract the poses of the desired topics from the BAG file

        # map<true_pose_topic,History<timestamp, SE3>>
        dict_history = ROSBag_Pose.extract_poses(bag=bag,
                                                 dict_topic_pose_body=dict_cfg["true_pose_topics"],
                                                 dict_senor_topic_pose=dict_cfg["true_pose_topics"],
                                                 num_messages=num_messages,
                                                 round_decimals=round_decimals,
                                                 dict_T_BODY_SENSOR=dict_T_BODY_SENSOR)
        dict_bsplines = dict()
        for true_pose_topic, hist in dict_history.items():
            if hist:
                bspline = BsplineSE3()
                if interp_type == TrajectoryInterpolationType.cubic:
                    bspline.feed_pose_history(hist_pose=hist, min_dt=min_dt, round_decimals=round_decimals)
                elif interp_type == TrajectoryInterpolationType.linear:
                    bspline.feed_pose_history(hist_pose=hist, uniform_timestamps=False)
                dict_bsplines[true_pose_topic] = bspline
        return dict_bsplines, dict_history


def main():
    # example: ROSBag_Pose2Ranges.py --bagfile ../test/sample_data//uwb_calib_a01_2023-08-31-21-05-46.bag --topic /d01/mavros/vision_pose/pose --cfg ../test/sample_data/config.yaml --verbose
    parser = argparse.ArgumentParser(
        description='ROSBag_TrueRelPoses: extract given pose topics and compute for each relative poses measurement a' +
                    'true relative pose (which can be perturbed), which overwrites to original message or is stored ' +
                    'under a new topic in the specified bag file.')
    parser.add_argument('--bagfile_in', help='input bag file', required=True)
    parser.add_argument('--bagfile_out', help='output bag file', required=True)
    parser.add_argument('--cfg',
                        help='YAML configuration file describing the setup: ' +
                             '{sensor_positions:{<id>:[x,y,z], ...}, sensor_orientations:{<id>:[w,x,y,z], ...}, ' +
                             'relpose_topics:{<id>:<topic_name>, ...}, true_pose_topics:{<id>:<topic_name>, ...},' +
                             ' new_relpose_topics:{<id>:<topic_name>, ...}}',
                        default="config.yaml", required=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--std_pos',
                        help='standard deviation of generated measurements: z_p = p + white_noise(std_p)',
                        default=0.0)
    parser.add_argument('--std_or',
                        help='standard deviation of generated measurements: z_R = R *R(white_noise(std_or))',
                        default=0.0)
    parser.add_argument('--use_header_timestamp', action='store_true',
                        help='overwrites the bag time with the header time stamp', default=False)
    parser.add_argument('--replace_with_new_topic', action='store_true',
                        help='removes relpose_topics if a new_relpose_topics was specified', default=False)
    parser.add_argument('--interpolation_type', help='Trajectory interpolation type',
                        choices=TrajectoryInterpolationType.list(),
                        default=str(TrajectoryInterpolationType.linear))
    parser.add_argument('--min_dt',
                        help='temporal displacement of cubic spline control points',
                        default=0.05)
    parser.add_argument('--pose_error_type', help='Covariance perturbation type (space) of relative pose measurements',
                        choices=EstimationErrorType.list(),
                        default=str(EstimationErrorType.type5))
    tp_start = time.time()
    args = parser.parse_args()

    if ROSBag_TrueRelPoses.extract(bagfile_in_name=args.bagfile_in,
                                   bagfile_out_name=args.bagfile_out,
                                   cfg=args.cfg,
                                   verbose=args.verbose,
                                   stddev_pos=float(args.std_pos),
                                   stddev_or=float(args.std_or),
                                   use_header_timestamp=args.use_header_timestamp,
                                   interp_type=TrajectoryInterpolationType(args.interpolation_type),
                                   min_dt=float(args.min_dt),
                                   pose_error_type=EstimationErrorType(args.pose_error_type)):
        print(" ")
        print("finished after [%s sec]\n" % str(time.time() - tp_start))
    else:
        print("failed! after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass



#!/usr/bin/env python
# Software License Agreement (GNU GPLv3  License)
#
# Copyright (c) 2023, Roland Jung (roland.jung@aau.at) , AAU, KPK, NAV
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# BASED ON: https://github.com/aau-cns/cnspy_rosbag2csv
# just install "pip install cnspy-rosbag2csv"
########################################################################################################################

import rosbag
import time
import os
import argparse
import yaml
import csv
from tqdm import tqdm
from numpy import linalg as LA
from spatialmath import UnitQuaternion

#from uwb_msgs.msg import TwoWayRangeStamped
from cnspy_trajectory.HistoryBuffer import get_key_from_value


class RelPose_ROSBag2CSV:
    def __init__(self):
        pass

    #
    # extract
    #
    @staticmethod
    def extract(bagfile_name, cfg, result_dir=None, fn_list=[], verbose=False, with_cov=False
                ):
        if not os.path.isfile(bagfile_name):
            print("RelPose_ROSBag2CSV: could not find file: %s" % bagfile_name)
            return False

        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("RelPose_ROSBag2CSV: could not find file: %s" % cfg)
            return False

        if verbose:
            print("RelPose_ROSBag2CSV:")
            print("* bagfile name: " + str(bagfile_name))
            print("* cfg: \t " + str(cfg))
            print("* filename_list: " + str(fn_list))

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "relpose_topics" not in dict_cfg:
                print("[relpose_topics] does not exist in fn=" + cfg)
                return False
            print("Read successful")

        if verbose:
            print("configuration contains:")
            print("relpose_topics:" + str(dict_cfg["relpose_topics"]))
            
        topic_list = dict_cfg["relpose_topics"].values()
        if len(topic_list) < 1:
            print("RelPose_ROSBag2CSV: no topics specified!")
            return False

        if fn_list:
            if len(topic_list) != len(fn_list):
                print("RelPose_ROSBag2CSV: topic_list and fn_list must have the same length!")
                return False
        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_name)
        except:
            if verbose:
                print("RelPose_ROSBag2CSV: Unexpected error!")

            return False

        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("RelPose_ROSBag2CSV: Unexpected error, bag file might be empty!")
            bag.close()
            return False

        ## create result dir:
        if result_dir is None:
            folder = str(bagfile_name).replace(".bag", "")
        else:
            folder = str(result_dir)

        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(folder))

        ## create csv file according to the topic names:
        dict_file_writers = dict()
        dict_header_written = dict()
        dict_csvfile_hdls = dict()
        idx = 0
        for id, topicName in dict_cfg["relpose_topics"].items():

            if topicName[0] != '/':
                print("RelPose_ROSBag2CSV: Not a proper topic name: %s (should start with /)" % topicName)
                continue

            if not fn_list:
                filename = str(folder + '/') + str.replace(topicName[1:], '/', '_') + '.csv'
            else:
                fn = fn_list[idx]
                [root, ext] = os.path.splitext(fn)
                [head, tail] = os.path.split(root)
                if ext:
                    filename = str(folder + '/') + tail + ext
                else:
                    filename = str(folder + '/') + tail + '.csv'

            csvfile = open(filename, 'w+')
            file_writer = csv.writer(csvfile, delimiter=',', lineterminator='\n')
            dict_file_writers[topicName] = file_writer
            dict_header_written[topicName] = False
            dict_csvfile_hdls[topicName] = csvfile

            if verbose:
                print("RelPose_ROSBag2CSV: creating csv file: %s " % filename)

            idx = idx + 1

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']
        for id, topicName in dict_cfg["relpose_topics"].items():
            found = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found = True

            if not found:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")

        if verbose:
            print("\nRelPose_ROSBag2CSV: num messages " + str(num_messages))

        round_decimals = 4
        cnt = 0
        ## extract the desired topics from the BAG file
        for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
            if topic in dict_cfg["relpose_topics"].values():
                if hasattr(msg, 'header') and hasattr(msg, 'poses'):  # STAMPED
                    file_writer = dict_file_writers[topic]

                    if not dict_header_written[topic]:
                        file_writer.writerow(RelPose_ROSBag2CSV.get_header(with_cov))
                        dict_header_written[topic] = True

                    ID1 = get_key_from_value(dict_cfg["relpose_topics"], topic)
                    timestamp = msg.header.stamp.to_sec()
                    for pose_id in msg.poses:
                        content = RelPose_ROSBag2CSV.to_csv_line(timestamp, ID1, pose_id, round_decimals, with_cov)
                        file_writer.writerow(content)
                        cnt += 1

        ## CLEANUP:
        # close all csv files
        for topicName in topic_list:
            dict_csvfile_hdls[topicName].close()

        # check if a topic was found by checking if the topic header was written
        for topicName in topic_list:
            if not dict_header_written[topicName]:
                print("\nRelPose_ROSBag2CSV: \n\tWARNING topic [" + str(topicName) + "] was not in bag-file")
                print("\tbag file [" + str(bagfile_name) + "] contains: ")
                # print(info_dict['topics'])
                for t in info_dict['topics']:
                    print(t['topic'])
                return False

        if verbose:
            print("\nRelPose_ROSBag2CSV.extract(): extracted [%d] msgs." % cnt)

        bag.close()
        return True

    #
    # extract_to_one
    #
    @staticmethod
    def extract_to_one(bagfile_name, cfg, fn, result_dir="", ext="csv", verbose=False, with_cov=False ) -> (bool, set):
        if not os.path.isfile(bagfile_name):
            print("RelPose_ROSBag2CSV: could not find file: %s" % bagfile_name)
            return False, set()

        cfg = os.path.abspath(cfg)
        if not os.path.isfile(cfg):
            print("RelPose_ROSBag2CSV: could not find file: %s" % cfg)
            return False, set()

        if verbose:
            print("RelPose_ROSBag2CSV:")
            print("* bagfile: " + str(bagfile_name))
            print("* cfg: \t " + str(cfg))
            print("* filename: " + str(fn))

        dict_cfg = None
        with open(cfg, "r") as yamlfile:
            dict_cfg = yaml.load(yamlfile, Loader=yaml.FullLoader)
            if "relpose_topics" not in dict_cfg:
                print("[relpose_topics] does not exist in fn=" + cfg)
                return False, set()
            print("Read successful")

        if verbose:
            print("configuration contains:")
            print("relpose_topics:" + str(dict_cfg["relpose_topics"]))

        topic_list = dict_cfg["relpose_topics"].values()
        if len(topic_list) < 1:
            print("RelPose_ROSBag2CSV: no topics specified!")
            return False, set()

        ## Open BAG file:
        try:
            bag = rosbag.Bag(bagfile_name)
        except:
            if verbose:
                print("RelPose_ROSBag2CSV: Unexpected error!")

            return False, set()

        info_dict = yaml.load(bag._get_yaml_info(), Loader=yaml.FullLoader)

        if info_dict is None or 'messages' not in info_dict:
            if verbose:
                print("RelPose_ROSBag2CSV: Unexpected error, bag file might be empty!")
            bag.close()
            return False, set()

        ## create result dir:
        if result_dir == "":
            folder = str(bagfile_name).replace(".bag", "")
        else:
            folder = result_dir

        folder = os.path.abspath(folder)
        try:  # else already exists
            os.makedirs(folder)
        except:
            pass

        if verbose:
            print("* result_dir: \t " + str(folder))

        if not fn:
          filename = str(folder + '/') + 'all_ranges.csv'
        else:
          filename = fn

        if verbose:
            print("* topic_list: \t " + str(topic_list))
            print("* filename_list: " + str(filename))

        for topicName in topic_list:
            if topicName[0] != '/':
                print("RelPose_ROSBag2CSV: Not a proper topic name: %s (should start with /)" % topicName)
                continue

        [root, ext] = os.path.splitext(filename)
        [head, tail] = os.path.split(root)
        if ext:
            filename = str(folder + '/') + tail + ext
        else:
            filename = str(folder + '/') + tail + '.csv'

        csvfile = open(filename, 'w+')
        file_writer = csv.writer(csvfile, delimiter=',', lineterminator='\n')
        file_writer.writerow(RelPose_ROSBag2CSV.get_header(with_cov))

        ## check if desired topics are in the bag file:
        num_messages = info_dict['messages']
        bag_topics = info_dict['topics']
        for topicName in topic_list:
            found = False
            for topic_info in bag_topics:
                if topic_info['topic'] == topicName:
                    found = True

            if not found:
                print("# WARNING: desired topic [" + str(topicName) + "] is not in bag file!")

        if verbose:
            print("\nRelPose_ROSBag2CSV: num messages " + str(num_messages))

        round_decimals = 4
        cnt = 0

        IDs = set()
        ## extract the desired topics from the BAG file
        for topic, msg, t in tqdm(bag.read_messages(), total=num_messages, unit="msgs"):
            if topic in topic_list:
                if hasattr(msg, 'header') and hasattr(msg, 'poses'):  # STAMPED
                    # HINT: conversions:

                    ID1 = get_key_from_value(dict_cfg["relpose_topics"], topic)
                    timestamp = msg.header.stamp.to_sec()
                    IDs.add(ID1)
                    for pose_id in msg.poses:
                        content = RelPose_ROSBag2CSV.to_csv_line(timestamp, ID1, pose_id, round_decimals, with_cov)
                        IDs.add(pose_id.id)
                        file_writer.writerow(content)
                        cnt += 1

        ## CLEANUP:
        csvfile.close()


        if verbose:
            print("\nRelPose_ROSBag2CSV.extract_to_one(): extracted [%d] msgs." % cnt)
            print("\nRelPose_ROSBag2CSV.extract_to_one(): IDs found: [%s]." % str(IDs))
        bag.close()
        return True, IDs

    @staticmethod
    def get_header(with_cov=False):
        if with_cov:
            return ["t", "ID1", "ID2", "tx", "ty", "tz", "qw", "qx", "qy", "qz", "range", "angle", 'pxx', 'pxy', 'pxz', 'pyy', 'pyz', 'pzz', 'qrr',
                    'qrp', 'qry', 'qpp', 'qpy', 'qyy']
        else:
            return ["t", "ID1", "ID2", "tx", "ty", "tz", "qw", "qx", "qy", "qz", "range", "angle"]

    @staticmethod
    def to_csv_line(timestamp, ID1, pose_id, round_decimals = 4, with_cov=False):
        range = LA.norm([pose_id.pose.position.x, pose_id.pose.position.y, pose_id.pose.position.z])
        q = UnitQuaternion(
            [pose_id.pose.orientation.w, pose_id.pose.orientation.x, pose_id.pose.orientation.y,
             pose_id.pose.orientation.z], norm=True).unit()
        [ang, vec] = q.angvec()
        ID2 = pose_id.id

        content = ["%f" % timestamp, str(ID1), str(ID2),
                   str(round(pose_id.pose.position.x, round_decimals)),
                   str(round(pose_id.pose.position.y, round_decimals)),
                   str(round(pose_id.pose.position.z, round_decimals)),
                   str(round(pose_id.pose.orientation.w, round_decimals)),
                   str(round(pose_id.pose.orientation.x, round_decimals)),
                   str(round(pose_id.pose.orientation.y, round_decimals)),
                   str(round(pose_id.pose.orientation.z, round_decimals)),
                   str(round(range, round_decimals)),
                   str(round(ang, round_decimals))]
        if with_cov:
            P = pose_id.covariance
            cov_tri_up = [P[0], P[1], P[2], P[7], P[8], P[14], P[21], P[22], P[23], P[28], P[29], P[35]]
            cov = []
            for i in cov_tri_up:
                cov.append(str(round(i, round_decimals)))
            content.extend(cov)

        return content

def main():
    # test3: python3 TWR_ROSbag2CS.py --bagfile ../test/example.bag --topics /a01/ranging /a02/ranging--verbose --filenames ranges.csv
    parser = argparse.ArgumentParser(
        description='RelPose_ROSBag2CSV: extract and store given topics of a rosbag into a CSV file')
    parser.add_argument('--bagfile', help='input bag file', default="not specified")
    parser.add_argument('--cfg', help='YAML configuration file describing the setup: {relpose_topics}', default="config.yaml", required=True)
    parser.add_argument('--filenames', nargs='*', help='csv filename of corresponding topic', default=[])
    parser.add_argument('--result_dir', help='directory to store results [otherwise bagfile name will be a directory]',
                        default='')
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--with_cov', action='store_true', default=False)
    tp_start = time.time()
    args = parser.parse_args()


    res = RelPose_ROSBag2CSV.extract(bagfile_name=args.bagfile,
                                 cfg=args.cfg,
                                 fn_list=args.filenames,
                                 result_dir=args.result_dir,
                                 verbose=args.verbose,
                                 with_cov=args.with_cov)

    if res:
        print(" ")
        print("finished after [%s sec]\n" % str(time.time() - tp_start))
    else:
        print("failed! after [%s sec]\n" % str(time.time() - tp_start))
    pass


if __name__ == "__main__":
    main()
    pass
